# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

try:
    from sklearn.preprocessing import normalize as sklearn_normalize
except ImportError:  # pragma: no cover
    sklearn_normalize = None

from ... import opcodes as OperandDef
from ...serialize import KeyField, StringField, Int32Field, BoolField
from ...tensor.operands import TensorOperand, TensorOperandMixin
from ...tensor.core import TensorOrder
from ...tensor.array_utils import as_same_device, device, sparse, issparse
from ...utils import recursive_tile
from ... import tensor as mt
from ..utils import check_array


class TensorNormalize(TensorOperand, TensorOperandMixin):
    _op_module_ = 'learn'
    _op_type_ = OperandDef.NORMALIZE

    _input = KeyField('input')
    _norm = StringField('norm')
    _axis = Int32Field('axis')
    _return_norm = BoolField('return_norm')
    # for test purpose
    _use_sklearn = BoolField('use_sklearn')

    def __init__(self, norm=None, axis=None, return_norm=None,
                 use_sklearn=None, dtype=None, gpu=None, **kw):
        super().__init__(_norm=norm, _axis=axis, _return_norm=return_norm,
                         _use_sklearn=use_sklearn, _dtype=dtype, _gpu=gpu, **kw)
        if self._use_sklearn is None:
            # force to use sklearn if not specified
            self._use_sklearn = True

    @property
    def input(self):
        return self._input

    @property
    def norm(self):
        return self._norm

    @property
    def axis(self):
        return self._axis

    @property
    def return_norm(self):
        return self._return_norm

    @property
    def use_sklearn(self):
        return self._use_sklearn

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input = self._inputs[0]

    @property
    def output_limit(self):
        return 2 if self._return_norm else 1

    def __call__(self, x, copy=True):
        x = check_array(x, accept_sparse=True,
                        estimator='the normalize function',
                        dtype=(np.float64, np.float32, np.float16))

        normed = None
        if not self._return_norm:
            res = self.new_tensor([x], shape=x.shape,
                                  order=x.order)
        else:
            kws = [
                {'shape': x.shape,
                 'order': x.order},
                {'shape': (x.shape[0] if self._axis == 1 else x.shape[1],),
                 'order': TensorOrder.C_ORDER}
            ]
            res, normed = self.new_tensors([x], kws=kws, output_limit=2)

        if not copy and self._axis == 1:
            # follow the behaviour of sklearn
            x.data = res.data

        if normed is None:
            return res
        return res, normed

    @classmethod
    def _tile_one_chunk(cls, op):
        outs = op.outputs
        chunk_op = op.copy().reset_key()
        kws = [
            {'shape': outs[0].shape,
             'order': outs[0].order,
             'index': (0, 0)}]
        if len(outs) == 2:
            kws.append({'shape': outs[1].shape,
                        'order': outs[1].order,
                        'index': (0,)})
        chunks = chunk_op.new_chunks([op.input.chunks[0]], kws=kws,
                                     output_limit=len(outs))

        tensor_kws = [
            {'shape': outs[0].shape,
             'order': outs[0].order,
             'chunks': [chunks[0]],
             'nsplits': tuple((s,) for s in outs[0].shape)
             }
        ]
        if len(outs) == 2:
            tensor_kws.append({'shape': outs[1].shape,
                               'order': outs[1].order,
                               'chunks': [chunks[1]],
                               'nsplits': tuple((s,) for s in outs[1].shape)
                               })

        new_op = op.copy()
        return new_op.new_tensors(op.inputs, kws=tensor_kws,
                                  output_limit=len(outs))

    @classmethod
    def _need_tile_into_chunks(cls, op):
        # if true, try to tile into chunks
        # whose implementation is based on sklearn itself
        x = op.input
        if op.gpu:  # pragma: no cover
            return False
        if x.issparse() and op.return_norm and op.norm in ('l1', 'l2'):
            # sklearn cannot handle
            return False
        if x.chunk_shape[op.axis] > 1:
            return False
        return True

    @classmethod
    def _tile_chunks(cls, op):
        assert op.input.chunk_shape[op.axis] == 1
        x = op.input
        axis = op.axis
        outs = op.outputs

        out_chunks = [], []
        for i, c in enumerate(x.chunks):
            chunk_op = op.copy().reset_key()
            kws = [
                {'shape': c.shape,
                 'order': c.order,
                 'index': c.index}
            ]
            if op.return_norm:
                kws.append({
                    'shape': (c.shape[1 - axis],),
                    'order': TensorOrder.C_ORDER,
                    'index': (i,),
                })
            chunks = chunk_op.new_chunks([c], kws=kws,
                                         output_limit=op.output_limit)
            out_chunks[0].append(chunks[0])
            if len(chunks) == 2:
                out_chunks[1].append(chunks[1])

        tensor_kws = [
            {'shape': outs[0].shape,
             'order': outs[0].order,
             'chunks': out_chunks[0],
             'nsplits': x.nsplits}
        ]
        if len(outs) == 2:
            tensor_kws.append({
                'shape': outs[1].shape,
                'order': outs[1].order,
                'chunks': out_chunks[1],
                'nsplits': (x.nsplits[1 - axis],)
            })
        new_op = op.copy()
        return new_op.new_tensors(op.inputs, kws=tensor_kws,
                                  output_limit=len(outs))

    @classmethod
    def tile(cls, op):
        x = op.input
        norm = op.norm
        axis = op.axis

        if len(x.chunks) == 1:
            return cls._tile_one_chunk(op)

        if cls._need_tile_into_chunks(op):
            return cls._tile_chunks(op)
        else:
            if norm == 'l1':
                norms = mt.abs(x).sum(axis=axis)
            elif norm == 'l2':
                norms = mt.sqrt((x ** 2).sum(axis=axis))
            else:
                assert norm == 'max'
                # sparse.max will still be a sparse,
                # force to convert to dense
                norms = mt.max(x, axis=axis).todense()
            norms = mt.where(mt.equal(norms, 0.0), 1.0, norms)
            if axis == 1:
                x = x / norms[:, mt.newaxis]
            else:
                x = x / norms[mt.newaxis, :]

            ret = [recursive_tile(x)]
            if op.return_norm:
                ret.append(recursive_tile(norms))

            new_op = op.copy()
            kws = [out.params for out in op.outputs]
            for i, r in enumerate(ret):
                kws[i]['chunks'] = r.chunks
                kws[i]['nsplits'] = r.nsplits
            return new_op.new_tensors(op.inputs, kws=kws)

    @classmethod
    def execute(cls, ctx, op):
        (x,), device_id, xp = as_same_device(
            [ctx[inp.key] for inp in op.inputs], device=op.device, ret_extra=True)
        axis = op.axis
        return_norm = op.return_norm
        norm = op.norm
        outs = op.outputs

        with device(device_id):
            if device_id < 0 and op.use_sklearn and sklearn_normalize is not None:
                # no GPU
                try:
                    if xp is sparse:
                        if axis == 0:
                            xm = x.raw.tocsc()
                        else:
                            xm = x.raw
                    else:
                        xm = x
                    ret = sklearn_normalize(xm, norm=norm, axis=axis,
                                            return_norm=return_norm)
                    normed = None
                    if return_norm:
                        ret, normed = ret
                    if issparse(ret):
                        ret = sparse.SparseNDArray(ret)
                    ctx[outs[0].key] = ret
                    if normed is not None:
                        ctx[outs[1].key] = normed
                    return
                except NotImplementedError:
                    pass

            # fall back
            if axis == 0:
                x = x.T

            if norm == 'l1':
                norms = xp.abs(x).sum(axis=1)
            elif norm == 'l2':
                norms = xp.sqrt((x ** 2).sum(axis=1))
            else:
                norms = xp.max(x, axis=1)
                if issparse(norms):
                    norms = norms.toarray()
            norms[norms == 0.0] = 1.0
            x = x / norms[:, np.newaxis]

            if axis == 0:
                x = x.T

            ctx[outs[0].key] = x
            if return_norm:
                ctx[outs[1].key] = norms


def normalize(X, norm='l2', axis=1, copy=True, return_norm=False):
    """
    Scale input vectors individually to unit norm (vector length).

    Parameters
    ----------
    X : {array-like, sparse matrix}, shape [n_samples, n_features]
        The data to normalize, element by element.
        scipy.sparse matrices should be in CSR format to avoid an
        un-necessary copy.

    norm : 'l1', 'l2', or 'max', optional ('l2' by default)
        The norm to use to normalize each non zero sample (or each non-zero
        feature if axis is 0).

    axis : 0 or 1, optional (1 by default)
        axis used to normalize the data along. If 1, independently normalize
        each sample, otherwise (if 0) normalize each feature.

    copy : boolean, optional, default True
        set to False to perform inplace row normalization and avoid a
        copy (if the input is already a tensor and if axis is 1).

    return_norm : boolean, default False
        whether to return the computed norms

    Returns
    -------
    X : {array-like, sparse matrix}, shape [n_samples, n_features]
        Normalized input X.

    norms : Tensor, shape [n_samples] if axis=1 else [n_features]
        A tensor of norms along given axis for X.
        When X is sparse, a NotImplementedError will be raised
        for norm 'l1' or 'l2'.

    See also
    --------
    Normalizer: Performs normalization using the ``Transformer`` API
        (e.g. as part of a preprocessing :class:`mars.learn.pipeline.Pipeline`).
    """
    if norm not in ('l1', 'l2', 'max'):
        raise ValueError(f"'{norm}' is not a supported norm")
    if axis not in (0, 1):
        raise ValueError(f"'{axis}' is not a supported axis")

    op = TensorNormalize(norm=norm, axis=axis, return_norm=return_norm,
                         dtype=np.dtype(np.float64))
    return op(X, copy=copy)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
import numbers

import numpy as np
from numpy.core.numeric import ComplexWarning
try:
    from sklearn.utils.validation import check_is_fitted
    from sklearn.exceptions import DataConversionWarning
except ImportError:  # pragma: no cover
    check_is_fitted = None
    DataConversionWarning = UserWarning

from ... import dataframe as md
from ... import tensor as mt
from ...core import ExecutableTuple
from ...lib.sparse import issparse
from ...tensor import Tensor
from .checks import check_non_negative_then_return_value, \
    assert_all_finite, AssertAllFinite


# ---------------------------------------------------------
# Original implementation is in `sklearn.utils.validation`.
# ---------------------------------------------------------

assert_all_finite = _assert_all_finite = assert_all_finite


def _num_samples(x):
    """Return number of samples in array-like x."""
    if hasattr(x, 'fit') and callable(x.fit):
        # Don't get num_samples from an ensembles length!
        raise TypeError(f'Expected sequence or array-like, got estimator {x}')
    if not hasattr(x, '__len__') and not hasattr(x, 'shape'):
        if hasattr(x, '__array__'):
            x = mt.asarray(x)
        else:
            raise TypeError(f"Expected sequence or array-like, got {type(x)}")
    if hasattr(x, 'shape'):
        if len(x.shape) == 0:
            if isinstance(x.op, AssertAllFinite):
                x = x.op.x
            if hasattr(x.op, 'data') and x.op.data is not None:
                x = np.asarray(x.op.data)
            raise TypeError(f"Singleton array {x!r} cannot be considered"
                            " a valid collection.")
        # Check that shape is returning an integer or default to len
        if isinstance(x.shape[0], numbers.Integral):
            return x.shape[0]
        elif np.isnan(x.shape[0]):
            return x.shape[0]
        else:
            return len(x)
    else:
        return len(x)


def check_consistent_length(*arrays, session=None, run_kwargs=None):
    """Check that all arrays have consistent first dimensions.

    Checks whether all objects in arrays have the same shape or length.

    Parameters
    ----------
    *arrays : list or tuple of input objects.
        Objects that will be checked for consistent length.
    """

    new_arrays = []
    lengths = []
    to_execute = []
    for X in arrays:
        if X is not None:
            n = _num_samples(X)
            if np.isnan(n):
                to_execute.append(X)
            new_arrays.append(X)
            lengths.append(n)
    # unknown length exists
    if len(to_execute) > 0:
        # update shape
        ExecutableTuple(to_execute).execute(session=session,
                                            **(run_kwargs or dict()))
        # get length again
        lengths = [_num_samples(X) for X in new_arrays]

    uniques = np.unique(lengths)
    if len(uniques) > 1:
        raise ValueError("Found input variables with inconsistent numbers of"
                         f" samples: {[int(length) for length in lengths]}")


def _make_indexable(iterable):
    """Ensure iterable supports indexing or convert to an indexable variant.

    Convert sparse matrices to csr and other non-indexable iterable to arrays.
    Let `None` and indexable objects (e.g. pandas dataframes) pass unchanged.

    Parameters
    ----------
    iterable : {list, dataframe, array, sparse} or None
        Object to be converted to an indexable iterable.
    """
    if issparse(iterable):
        return mt.tensor(iterable)
    elif hasattr(iterable, "iloc"):
        if iterable.ndim == 1:
            return md.Series(iterable)
        else:
            return md.DataFrame(iterable)
    elif hasattr(iterable, "__getitem__"):
        return mt.tensor(iterable)
    elif iterable is None:
        return iterable
    return mt.tensor(iterable)


def indexable(*iterables, session=None, run_kwargs=None):
    """Make arrays indexable for cross-validation.

    Checks consistent length, passes through None, and ensures that everything
    can be indexed by converting sparse matrices to csr and converting
    non-interable objects to arrays.

    Parameters
    ----------
    *iterables : lists, dataframes, arrays, sparse matrices
        List of objects to ensure sliceability.
    """
    result = [_make_indexable(X) for X in iterables]
    check_consistent_length(*result, session=session,
                            run_kwargs=run_kwargs)
    return result


def _ensure_no_complex_data(array):
    if hasattr(array, 'dtype') and array.dtype is not None \
            and hasattr(array.dtype, 'kind') and array.dtype.kind == "c":
        raise ValueError(f"Complex data not supported\n{array}\n")


def _ensure_sparse_format(spmatrix, accept_sparse, dtype, copy,
                          force_all_finite, accept_large_sparse):
    """Convert a sparse matrix to a given format.

    Checks the sparse format of spmatrix and converts if necessary.

    Parameters
    ----------
    spmatrix : scipy sparse matrix
        Input to validate and convert.

    accept_sparse : string, boolean or list/tuple of strings
        String[s] representing allowed sparse matrix formats ('csc',
        'csr', 'coo', 'dok', 'bsr', 'lil', 'dia'). If the input is sparse but
        not in the allowed format, it will be converted to the first listed
        format. True allows the input to be any format. False means
        that a sparse matrix input will raise an error.

    dtype : string, type or None
        Data type of result. If None, the dtype of the input is preserved.

    copy : boolean
        Whether a forced copy will be triggered. If copy=False, a copy might
        be triggered by a conversion.

    force_all_finite : boolean or 'allow-nan', (default=True)
        Whether to raise an error on np.inf and np.nan in X. The possibilities
        are:

        - True: Force all values of X to be finite.
        - False: accept both np.inf and np.nan in X.
        - 'allow-nan': accept only np.nan values in X. Values cannot be
          infinite.

    Returns
    -------
    spmatrix_converted : scipy sparse matrix.
        Matrix that is ensured to have an allowed type.
    """
    if dtype is None:
        dtype = spmatrix.dtype

    changed_format = False

    if isinstance(accept_sparse, str):
        accept_sparse = [accept_sparse]

    # Indices dtype validation
    # _check_large_sparse(spmatrix, accept_large_sparse)

    if accept_sparse is False:
        raise TypeError('A sparse tensor was passed, but dense '
                        'data is required. Use X.todense() to '
                        'convert to a dense tensor.')
    elif isinstance(accept_sparse, (list, tuple)):
        if len(accept_sparse) == 0:
            raise ValueError("When providing 'accept_sparse' "
                             "as a tuple or list, it must contain at "
                             "least one string value.")
        # # ensure correct sparse format
        # if spmatrix.format not in accept_sparse:
        #     # create new with correct sparse
        #     spmatrix = spmatrix.asformat(accept_sparse[0])
        #     changed_format = True
    elif accept_sparse is not True:
        # any other type
        raise ValueError("Parameter 'accept_sparse' should be a string, "
                         "boolean or list of strings. You provided "
                         f"'accept_sparse={accept_sparse}'.")

    if dtype != spmatrix.dtype:
        # convert dtype
        spmatrix = spmatrix.astype(dtype)
    elif copy and not changed_format:
        # force copy
        spmatrix = spmatrix.copy()

    if force_all_finite:
        spmatrix = assert_all_finite(
            spmatrix, allow_nan=force_all_finite == 'allow-nan', check_only=False)

    return spmatrix


def check_array(array, accept_sparse=False, accept_large_sparse=True,
                dtype="numeric", order=None, copy=False, force_all_finite=True,
                ensure_2d=True, allow_nd=False, ensure_min_samples=1,
                ensure_min_features=1, estimator=None) -> Tensor:

    """Input validation on a tensor, list, sparse matrix or similar.

    By default, the input is checked to be a non-empty 2D array containing
    only finite values. If the dtype of the tensor is object, attempt
    converting to float, raising on failure.

    Parameters
    ----------
    array : object
        Input object to check / convert.

    accept_sparse : string, boolean or list/tuple of strings (default=False)
        String[s] representing allowed sparse matrix formats, such as 'csc',
        'csr', etc. If the input is sparse but not in the allowed format,
        it will be converted to the first listed format. True allows the input
        to be any format. False means that a sparse matrix input will
        raise an error.

    accept_large_sparse : bool (default=True)
        If a CSR, CSC, COO or BSR sparse matrix is supplied and accepted by
        accept_sparse, accept_large_sparse=False will cause it to be accepted
        only if its indices are stored with a 32-bit dtype.

    dtype : string, type, list of types or None (default="numeric")
        Data type of result. If None, the dtype of the input is preserved.
        If "numeric", dtype is preserved unless array.dtype is object.
        If dtype is a list of types, conversion on the first type is only
        performed if the dtype of the input is not in the list.

    order : 'F', 'C' or None (default=None)
        Whether a tenor will be forced to be fortran or c-style.
        When order is None (default), then if copy=False, nothing is ensured
        about the memory layout of the output tensor; otherwise (copy=True)
        the memory layout of the returned tensor is kept as close as possible
        to the original tensor.

    copy : boolean (default=False)
        Whether a forced copy will be triggered. If copy=False, a copy might
        be triggered by a conversion.

    force_all_finite : boolean or 'allow-nan', (default=True)
        Whether to raise an error on np.inf and np.nan in tensor. The
        possibilities are:

        - True: Force all values of tensor to be finite.
        - False: accept both np.inf and np.nan in tensor.
        - 'allow-nan': accept only np.nan values in tensor. Values cannot
          be infinite.

        For object dtyped data, only np.nan is checked and not np.inf.

    ensure_2d : boolean (default=True)
        Whether to raise a value error if tensor is not 2D.

    allow_nd : boolean (default=False)
        Whether to allow tensor.ndim > 2.

    ensure_min_samples : int (default=1)
        Make sure that the tensor has a minimum number of samples in its first
        axis (rows for a 2D tensor). Setting to 0 disables this check.

    ensure_min_features : int (default=1)
        Make sure that the 2D tensor has some minimum number of features
        (columns). The default value of 1 rejects empty datasets.
        This check is only enforced when the input data has effectively 2
        dimensions or is originally 1D and ``ensure_2d`` is True. Setting to 0
        disables this check.

    estimator : str or estimator instance (default=None)
        If passed, include the name of the estimator in warning messages.

    Returns
    -------
    array_converted : object
        The converted and validated tensor.
    """

    # store whether originally we wanted numeric dtype
    dtype_numeric = isinstance(dtype, str) and dtype == "numeric"

    dtype_orig = getattr(array, "dtype", None)
    if not hasattr(dtype_orig, 'kind'):
        # not a data type (e.g. a column named dtype in a pandas DataFrame)
        dtype_orig = None

    if dtype_numeric:
        if dtype_orig is not None and dtype_orig.kind == "O":
            # if input is object, convert to float.
            dtype = np.float64
        else:
            dtype = None

    if isinstance(dtype, (list, tuple)):
        if dtype_orig is not None and dtype_orig in dtype:
            # no dtype conversion required
            dtype = None
        else:
            # dtype conversion required. Let's select the first element of the
            # list of accepted types.
            dtype = dtype[0]

    if force_all_finite not in (True, False, 'allow-nan'):
        raise ValueError('force_all_finite should be a bool or "allow-nan"'
                         f'. Got {force_all_finite!r} instead')

    if estimator is not None:
        if isinstance(estimator, str):
            estimator_name = estimator
        else:
            estimator_name = estimator.__class__.__name__
    else:
        estimator_name = "Estimator"
    context = f" by {estimator_name}" if estimator is not None else ""

    if (hasattr(array, 'issparse') and array.issparse()) or issparse(array):
        _ensure_no_complex_data(array)
        array = mt.asarray(array)
        array = _ensure_sparse_format(array, accept_sparse=accept_sparse,
                                      dtype=dtype, copy=copy,
                                      force_all_finite=force_all_finite,
                                      accept_large_sparse=accept_large_sparse)
    else:
        # If np.array(..) gives ComplexWarning, then we convert the warning
        # to an error. This is needed because specifying a non complex
        # dtype to the function converts complex to real dtype,
        # thereby passing the test made in the lines following the scope
        # of warnings context manager.
        with warnings.catch_warnings():
            try:
                warnings.simplefilter('error', ComplexWarning)
                array = mt.asarray(array, dtype=dtype, order=order)
            except ComplexWarning:
                raise ValueError(f"Complex data not supported\n{array}\n")

        # It is possible that the np.array(..) gave no warning. This happens
        # when no dtype conversion happened, for example dtype = None. The
        # result is that np.array(..) produces an array of complex dtype
        # and we need to catch and raise exception for such cases.
        _ensure_no_complex_data(array)

        if ensure_2d:
            # If input is scalar raise error
            if array.ndim == 0:
                raise ValueError(
                    f"Expected 2D array, got scalar array instead:\narray={array}.\n"
                    "Reshape your data either using array.reshape(-1, 1) if "
                    "your data has a single feature or array.reshape(1, -1) "
                    "if it contains a single sample.")
            # If input is 1D raise error
            if array.ndim == 1:
                raise ValueError(
                    f"Expected 2D array, got 1D array instead:\narray={array}.\n"
                    "Reshape your data either using array.reshape(-1, 1) if "
                    "your data has a single feature or array.reshape(1, -1) "
                    "if it contains a single sample.")

        # in the future np.flexible dtypes will be handled like object dtypes
        if dtype_numeric and np.issubdtype(array.dtype, np.flexible):
            warnings.warn(
                "Beginning in version 0.22, arrays of bytes/strings will be "
                "converted to decimal numbers if dtype='numeric'. "
                "It is recommended that you convert the array to "
                "a float dtype before using it in scikit-learn, "
                "for example by using "
                "your_array = your_array.astype(np.float64).",
                FutureWarning)

        # make sure we actually converted to numeric:
        if dtype_numeric and array.dtype.kind == "O":
            array = array.astype(np.float64)
        if not allow_nd and array.ndim >= 3:
            raise ValueError("Found array with dim %d. %s expected <= 2."
                             % (array.ndim, estimator_name))
        if force_all_finite:
            array = _assert_all_finite(
                array, allow_nan=force_all_finite == 'allow-nan', check_only=False)

    if ensure_min_samples > 0:
        n_samples = _num_samples(array)
        if n_samples < ensure_min_samples:
            raise ValueError("Found array with %d sample(s) (shape=%s) while a"
                             " minimum of %d is required%s."
                             % (n_samples, array.shape, ensure_min_samples,
                                context))

    if ensure_min_features > 0 and array.ndim == 2:
        n_features = array.shape[1]
        if n_features < ensure_min_features:
            raise ValueError("Found array with %d feature(s) (shape=%s) while"
                             " a minimum of %d is required%s."
                             % (n_features, array.shape, ensure_min_features,
                                context))

    if copy:
        array = mt.array(array, dtype=dtype, order=order)

    return array


def check_X_y(X, y, accept_sparse=False, accept_large_sparse=True,
              dtype="numeric", order=None, copy=False, force_all_finite=True,
              ensure_2d=True, allow_nd=False, multi_output=False,
              ensure_min_samples=1, ensure_min_features=1, y_numeric=False,
              estimator=None):
    """Input validation for standard estimators.

    Checks X and y for consistent length, enforces X to be 2D and y 1D. By
    default, X is checked to be non-empty and containing only finite values.
    Standard input checks are also applied to y, such as checking that y
    does not have np.nan or np.inf targets. For multi-label y, set
    multi_output=True to allow 2D and sparse y. If the dtype of X is
    object, attempt converting to float, raising on failure.

    Parameters
    ----------
    X : tensor, list or sparse tensor
        Input data.

    y : tensor, list or sparse tensor
        Labels.

    accept_sparse : string, boolean or list of string (default=False)
        String[s] representing allowed sparse matrix formats, such as 'csc',
        'csr', etc. If the input is sparse but not in the allowed format,
        it will be converted to the first listed format. True allows the input
        to be any format. False means that a sparse matrix input will
        raise an error.

    accept_large_sparse : bool (default=True)
        If a CSR, CSC, COO or BSR sparse matrix is supplied and accepted by
        accept_sparse, accept_large_sparse will cause it to be accepted only
        if its indices are stored with a 32-bit dtype.

    dtype : string, type, list of types or None (default="numeric")
        Data type of result. If None, the dtype of the input is preserved.
        If "numeric", dtype is preserved unless array.dtype is object.
        If dtype is a list of types, conversion on the first type is only
        performed if the dtype of the input is not in the list.

    order : 'F', 'C' or None (default=None)
        Whether an array will be forced to be fortran or c-style.

    copy : boolean (default=False)
        Whether a forced copy will be triggered. If copy=False, a copy might
        be triggered by a conversion.

    force_all_finite : boolean or 'allow-nan', (default=True)
        Whether to raise an error on np.inf and np.nan in X. This parameter
        does not influence whether y can have np.inf or np.nan values.
        The possibilities are:

        - True: Force all values of X to be finite.
        - False: accept both np.inf and np.nan in X.
        - 'allow-nan': accept only np.nan values in X. Values cannot be
          infinite.

    ensure_2d : boolean (default=True)
        Whether to raise a value error if X is not 2D.

    allow_nd : boolean (default=False)
        Whether to allow X.ndim > 2.

    multi_output : boolean (default=False)
        Whether to allow 2D y (array or sparse matrix). If false, y will be
        validated as a vector. y cannot have np.nan or np.inf values if
        multi_output=True.

    ensure_min_samples : int (default=1)
        Make sure that X has a minimum number of samples in its first
        axis (rows for a 2D array).

    ensure_min_features : int (default=1)
        Make sure that the 2D array has some minimum number of features
        (columns). The default value of 1 rejects empty datasets.
        This check is only enforced when X has effectively 2 dimensions or
        is originally 1D and ``ensure_2d`` is True. Setting to 0 disables
        this check.

    y_numeric : boolean (default=False)
        Whether to ensure that y has a numeric type. If dtype of y is object,
        it is converted to float64. Should only be used for regression
        algorithms.

    estimator : str or estimator instance (default=None)
        If passed, include the name of the estimator in warning messages.

    Returns
    -------
    X_converted : object
        The converted and validated X.

    y_converted : object
        The converted and validated y.
    """
    if y is None:
        raise ValueError('y cannot be None')

    X = check_array(X, accept_sparse=accept_sparse,
                    accept_large_sparse=accept_large_sparse,
                    dtype=dtype, order=order, copy=copy,
                    force_all_finite=force_all_finite,
                    ensure_2d=ensure_2d, allow_nd=allow_nd,
                    ensure_min_samples=ensure_min_samples,
                    ensure_min_features=ensure_min_features,
                    estimator=estimator)
    if multi_output:
        y = check_array(y, True, force_all_finite=True, ensure_2d=False,
                        dtype=None)
    else:
        y = column_or_1d(y, warn=True)
        y = _assert_all_finite(y, check_only=False)
    if y_numeric and y.dtype.kind == 'O':
        y = y.astype(np.float64)

    check_consistent_length(X, y)

    return X, y


def check_non_negative(X, whom):
    """
    Check if there is any negative value in a tensor.

    Parameters
    ----------
    X : array-like or sparse matrix
        Input data.

    whom : string
        Who passed X to this function.
    """
    return check_non_negative_then_return_value(X, X, whom)


def column_or_1d(y, warn=False):
    """ Ravel column or 1d numpy array, else raises an error

    Parameters
    ----------
    y : array-like

    warn : boolean, default False
       To control display of warnings.

    Returns
    -------
    y : array

    """
    y = mt.tensor(y)
    shape = y.shape
    if len(shape) == 1:
        return mt.ravel(y)
    if len(shape) == 2 and shape[1] == 1:
        if warn:
            warnings.warn("A column-vector y was passed when a 1d array was"
                          " expected. Please change the shape of y to "
                          "(n_samples, ), for example using ravel().",
                          DataConversionWarning, stacklevel=2)
        return mt.ravel(y)

    raise ValueError(f"bad input shape {shape}")


check_is_fitted = check_is_fitted


def _check_sample_weight(sample_weight, X, dtype=None):
    """Validate sample weights.

    Note that passing sample_weight=None will output an array of ones.
    Therefore, in some cases, you may want to protect the call with:
    if sample_weight is not None:
        sample_weight = _check_sample_weight(...)

    Parameters
    ----------
    sample_weight : {ndarray, Number or None}, shape (n_samples,)
       Input sample weights.

    X : nd-array, list or sparse matrix
        Input data.

    dtype: dtype
       dtype of the validated `sample_weight`.
       If None, and the input `sample_weight` is an array, the dtype of the
       input is preserved; otherwise an array with the default numpy dtype
       is be allocated.  If `dtype` is not one of `float32`, `float64`,
       `None`, the output will be of dtype `float64`.

    Returns
    -------
    sample_weight : ndarray, shape (n_samples,)
       Validated sample weight. It is guaranteed to be "C" contiguous.
    """
    n_samples = _num_samples(X)

    if dtype is not None and dtype not in [np.float32, np.float64]:
        dtype = np.float64

    if sample_weight is None or isinstance(sample_weight, numbers.Number):
        if sample_weight is None:
            sample_weight = mt.ones(n_samples, dtype=dtype)
        else:
            sample_weight = mt.full(n_samples, sample_weight,
                                    dtype=dtype)
    else:
        if dtype is None:
            dtype = [np.float64, np.float32]
        sample_weight = check_array(
            sample_weight, accept_sparse=False, ensure_2d=False, dtype=dtype,
            order="C"
        )
        if sample_weight.ndim != 1:
            raise ValueError("Sample weights must be 1D array or scalar")

        if sample_weight.shape != (n_samples,):
            raise ValueError(f"sample_weight.shape == {sample_weight.shape}, "
                             f"expected {(n_samples,)}!")
    return sample_weight



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import operator
import logging
import sys
import time
from collections import defaultdict
from enum import Enum

from .. import promise
from ..config import options
from ..errors import DependencyMissing, ExecutionInterrupted, WorkerDead
from ..serialize import dataserializer
from ..scheduler.chunkmeta import WorkerMeta
from ..utils import log_unhandled, build_exc_info
from .events import EventContext, EventCategory, EventLevel, ProcedureEventType
from .storage import DataStorageDevice
from .utils import WorkerActor

logger = logging.getLogger(__name__)


class ReceiveStatus(Enum):
    NOT_STARTED = 0
    PENDING = 1
    RECEIVING = 2
    RECEIVED = 3
    ERROR = 4


class EndpointTransferState(object):
    """
    Structure providing transfer status in an endpoint
    """
    __slots__ = 'parts', 'total_size', 'keys', 'end_marks', 'send_future'

    def __init__(self):
        self.reset()
        self.send_future = None

    def reset(self):
        self.parts = []
        self.total_size = 0
        self.keys = []
        self.end_marks = []


class SenderActor(WorkerActor):
    """
    Actor handling sending data to ReceiverActors in other workers
    """
    def __init__(self):
        super().__init__()
        self._dispatch_ref = None
        self._events_ref = None

    def post_create(self):
        from .dispatcher import DispatchActor
        from .events import EventsActor

        super().post_create()

        self._events_ref = self.ctx.actor_ref(EventsActor.default_uid())
        if not self.ctx.has_actor(self._events_ref):
            self._events_ref = None

        self._dispatch_ref = self.promise_ref(DispatchActor.default_uid())
        self._dispatch_ref.register_free_slot(self.uid, 'sender')

    @promise.reject_on_exception
    @log_unhandled
    def send_data(self, session_id, chunk_keys, target_endpoints, ensure_cached=True,
                  compression=None, block_size=None, pin_token=None, timeout=None, callback=None):
        """
        Send data to other workers
        :param session_id: session id
        :param chunk_keys: chunks to be sent
        :param target_endpoints: endpoints to receive this chunk
        :param ensure_cached: if True, make sure the data is in the shared storage of the target worker
        :param compression: compression type when transfer in network
        :param block_size: size of data block
        :param pin_token: token to pin the data
        :param timeout: timeout of data sending
        :param callback: promise callback
        """
        chunk_keys = list(chunk_keys)
        target_endpoints = list(target_endpoints)
        block_size = block_size or options.worker.transfer_block_size

        data_sizes = self.storage_client.get_data_sizes(session_id, chunk_keys)
        if any(s is None for s in data_sizes):
            raise DependencyMissing('Dependencies %r not met when sending.'
                                    % [k for k, s in zip(chunk_keys, data_sizes) if s is None])
        if options.vineyard.enabled:
            compression = None
        else:
            compression = compression or dataserializer.CompressType(options.worker.transfer_compression)

        wait_refs = []
        addrs_to_chunks = dict()
        keys_to_readers = dict()
        receiver_refs = []
        receiver_manager_ref_dict = dict(
            (ep, self.promise_ref(ReceiverManagerActor.default_uid(), address=ep))
            for ep in target_endpoints)

        @log_unhandled
        def _create_local_readers():
            promises = []
            for k in chunk_keys:
                promises.append(self.storage_client.create_reader(
                    session_id, k, source_devices, packed=True, packed_compression=compression)
                    .then(functools.partial(operator.setitem, keys_to_readers, k)))
            return promise.all_(promises)

        @log_unhandled
        def _create_remote_writers():
            nonlocal data_sizes
            create_write_promises = []
            data_sizes = self.storage_client.get_data_sizes(session_id, chunk_keys)
            for ref in receiver_manager_ref_dict.values():
                # register transfer actions
                create_write_promises.append(
                    ref.create_data_writers(
                        session_id, chunk_keys, data_sizes, self.ref(), ensure_cached=ensure_cached,
                        timeout=timeout, pin_token=pin_token, _timeout=timeout, _promise=True
                    ).then(_handle_created)
                )
            return promise.all_(create_write_promises).then(lambda *_: keys_to_readers)

        @log_unhandled
        def _handle_created(ref, statuses):
            manager_ref = receiver_manager_ref_dict[ref.address]
            # filter out endpoints already transferred or already started transfer
            keys_receiving, keys_to_receive = [], []
            for k, status in zip(chunk_keys, statuses):
                if status == ReceiveStatus.RECEIVING:
                    keys_receiving.append(k)
                elif status is None:
                    keys_to_receive.append(k)

            if keys_to_receive:
                addrs_to_chunks[ref.address] = keys_to_receive
                receiver_refs.append(self.promise_ref(ref))
            if keys_receiving:
                wait_refs.append(manager_ref.add_keys_callback(
                    session_id, keys_receiving, _timeout=timeout, _promise=True))

        @log_unhandled
        def _finalize(*_):
            self._dispatch_ref.register_free_slot(self.uid, 'sender', _tell=True)
            self.tell_promise(callback, data_sizes)

        @log_unhandled
        def _handle_rejection(*exc):
            logger.exception('Transfer chunks %r to %r failed', chunk_keys, target_endpoints, exc_info=exc)
            self._dispatch_ref.register_free_slot(self.uid, 'sender', _tell=True)

            for reader in keys_to_readers.values():
                reader.close()
            keys_to_readers.clear()

            for ref in receiver_refs:
                ref.cancel_receive(session_id, addrs_to_chunks[ref.address], _tell=True, _wait=False)
            self.tell_promise(callback, *exc, _accept=False)

        try:
            if options.vineyard.enabled:
                in_memory_device = DataStorageDevice.VINEYARD
            else:
                in_memory_device = DataStorageDevice.SHARED_MEMORY
            source_devices = [in_memory_device, DataStorageDevice.DISK]  # pragma: no cover
            _create_local_readers().then(_create_remote_writers) \
                .then(lambda *_: self._compress_and_send(
                    session_id, addrs_to_chunks, receiver_refs, keys_to_readers,
                    block_size=block_size, timeout=timeout,
                )) \
                .then(lambda *_: promise.all_(wait_refs)) \
                .then(_finalize, _handle_rejection)
        except:  # noqa: E722
            _handle_rejection(*sys.exc_info())
            return

    @log_unhandled
    def _compress_and_send(self, session_id, addrs_to_chunks, receiver_refs, keys_to_readers,
                           block_size, timeout=None):
        """
        Compress and send data to receivers in chunked manner
        :param session_id: session id
        :param addrs_to_chunks: dict mapping endpoints to chunks to send
        :param receiver_refs: refs to send data to
        """
        # collect data targets
        chunks_to_addrs = defaultdict(set)
        for addr, chunks in addrs_to_chunks.items():
            for key in chunks:
                chunks_to_addrs[key].add(addr)
        all_chunk_keys = sorted(chunks_to_addrs.keys(), key=lambda k: len(chunks_to_addrs[k]))
        addr_statuses = dict((k, EndpointTransferState()) for k in addrs_to_chunks.keys())
        addr_to_refs = dict((ref.address, ref) for ref in receiver_refs)

        # start compress and send data into targets
        logger.debug('Data writer for chunks %r allocated at targets, start transmission', all_chunk_keys)

        # filter out endpoints we need to send to
        try:
            if not receiver_refs:
                self._dispatch_ref.register_free_slot(self.uid, 'sender', _tell=True, _wait=False)
                return
            cur_key_id = 0
            cur_key = all_chunk_keys[cur_key_id]
            cur_reader = keys_to_readers[cur_key]
            with EventContext(self._events_ref, EventCategory.PROCEDURE, EventLevel.NORMAL,
                              ProcedureEventType.NETWORK, self.uid):
                while cur_key_id < len(all_chunk_keys):
                    # read a data part from reader we defined above
                    pool = cur_reader.get_io_pool()
                    next_part = pool.submit(cur_reader.read, block_size).result()
                    file_eof = len(next_part) < block_size

                    for addr in chunks_to_addrs[cur_key]:
                        addr_status = addr_statuses[addr]
                        addr_status.parts.append(next_part)
                        addr_status.keys.append(cur_key)
                        addr_status.total_size += len(next_part)
                        addr_status.end_marks.append(file_eof)

                        if addr_status.total_size >= block_size:
                            if addr_status.send_future:
                                addr_status.send_future.result(timeout=timeout)
                            addr_status.send_future = addr_to_refs[addr].receive_data_part(
                                session_id, addr_status.keys, addr_status.end_marks,
                                *addr_status.parts, _wait=False)
                            addr_status.reset()

                    # when some part goes to end, move to the next chunk
                    if file_eof:
                        cur_reader.close()
                        cur_key_id += 1
                        if cur_key_id < len(all_chunk_keys):
                            # still some chunks left unhandled
                            cur_key = all_chunk_keys[cur_key_id]
                            cur_reader = keys_to_readers[cur_key]
                        else:
                            # all chunks handled
                            for addr, addr_status in addr_statuses.items():
                                if addr_status.send_future:
                                    addr_status.send_future.result(timeout=timeout)
                                if addr_status.parts:
                                    # send remaining chunks
                                    addr_status.end_marks[-1] = True
                                    addr_status.send_future = addr_to_refs[addr].receive_data_part(
                                        session_id, addr_status.keys, addr_status.end_marks,
                                        *addr_status.parts, _wait=False)
                                addr_status.reset()
                            for addr_status in addr_statuses.values():
                                if addr_status.send_future:
                                    addr_status.send_future.result(timeout=timeout)
                                addr_status.reset()

        except:  # noqa: E722
            for ref in receiver_refs:
                ref.cancel_receive(session_id, addrs_to_chunks[ref.address], _tell=True, _wait=False)
            raise
        finally:
            for reader in keys_to_readers.values():
                reader.close()


class ReceiverDataMeta(object):
    __slots__ = 'start_time', 'chunk_size', 'source_address',\
                'status', 'transfer_event_id', 'receiver_worker_uid', \
                'callback_ids', 'callback_args', 'callback_kwargs'

    def __init__(self, start_time=None, chunk_size=None, source_address=None,
                 transfer_event_id=None, receiver_worker_uid=None, status=None,
                 callback_ids=None, callback_args=None, callback_kwargs=None):
        self.start_time = start_time
        self.chunk_size = chunk_size
        self.source_address = source_address
        self.status = status
        self.transfer_event_id = transfer_event_id
        self.receiver_worker_uid = receiver_worker_uid
        self.callback_ids = callback_ids or []
        self.callback_args = callback_args or ()
        self.callback_kwargs = callback_kwargs or {}

    def update(self, **kwargs):
        kwargs['callback_ids'] = list(set(self.callback_ids) | set(kwargs.get('callback_ids') or ()))
        for k, v in kwargs.items():
            setattr(self, k, v)


class ReceiverManagerActor(WorkerActor):
    def __init__(self):
        super().__init__()
        self._data_metas = dict()
        self._max_callback_id = 0
        self._callback_id_to_callbacks = dict()
        self._callback_id_to_keys = dict()

        self._dispatch_ref = None

    def post_create(self):
        super().post_create()

        from .dispatcher import DispatchActor
        self._dispatch_ref = self.promise_ref(DispatchActor.default_uid())

    def _update_data_meta(self, session_id, data_key, **kwargs):
        try:
            self._data_metas[(session_id, data_key)].update(**kwargs)
        except KeyError:
            self._data_metas[(session_id, data_key)] = ReceiverDataMeta(**kwargs)

    @promise.reject_on_exception
    @log_unhandled
    def create_data_writers(self, session_id, data_keys, data_sizes, sender_ref,
                            ensure_cached=True, pin_token=None, timeout=0,
                            use_promise=True, callback=None):
        sender_address = None if sender_ref is None else sender_ref.address
        logger.debug('Begin creating transmission data writer for chunks %r from %s',
                     data_keys, sender_address)
        data_locations = dict(zip(
            data_keys, self.storage_client.get_data_locations(session_id, data_keys)))
        keys_to_fetch = []
        sizes_to_fetch = []
        statuses = []

        slot_ref = self.promise_ref(self._dispatch_ref.get_hash_slot('receiver', repr(data_keys)))
        for chunk_key, data_size in zip(data_keys, data_sizes):
            session_chunk_key = (session_id, chunk_key)

            try:
                data_meta = self._data_metas[session_chunk_key]
            except KeyError:
                data_meta = self._data_metas[session_chunk_key] = \
                    ReceiverDataMeta(chunk_size=data_size, source_address=sender_address,
                                     status=ReceiveStatus.NOT_STARTED)

            if data_locations.get(chunk_key):
                data_meta.status = ReceiveStatus.RECEIVED
                statuses.append(ReceiveStatus.RECEIVED)
                self._update_data_meta(session_id, chunk_key, status=ReceiveStatus.RECEIVED)
                continue
            elif data_meta.status == ReceiveStatus.RECEIVING:
                # data transfer already started
                logger.debug('Chunk %s already started transmission', chunk_key)
                statuses.append(ReceiveStatus.RECEIVING)
                continue
            elif data_meta.status == ReceiveStatus.RECEIVED:
                data_meta = self._data_metas[session_chunk_key] = \
                    ReceiverDataMeta(chunk_size=data_size, source_address=sender_address,
                                     status=ReceiveStatus.NOT_STARTED)

            data_meta.start_time = time.time()
            data_meta.receiver_worker_uid = slot_ref.uid
            data_meta.source_address = sender_address
            data_meta.status = ReceiveStatus.RECEIVING

            self._update_data_meta(session_id, chunk_key, chunk_size=data_size,
                                   source_address=sender_address, status=ReceiveStatus.RECEIVING)
            keys_to_fetch.append(chunk_key)
            sizes_to_fetch.append(data_size)
            statuses.append(None)  # this notifies the sender to transmit data

        if use_promise:
            if keys_to_fetch:
                slot_ref.create_data_writers(
                    session_id, keys_to_fetch, sizes_to_fetch, sender_ref, ensure_cached=ensure_cached,
                    timeout=timeout, pin_token=pin_token, use_promise=use_promise, _promise=True) \
                    .then(lambda *_: self.tell_promise(callback, slot_ref, statuses))
            else:
                self.tell_promise(callback, slot_ref, statuses)
        else:
            slot_ref.create_data_writers(
                session_id, keys_to_fetch, sizes_to_fetch, sender_ref, ensure_cached=ensure_cached,
                timeout=timeout, pin_token=pin_token, use_promise=use_promise)
            return slot_ref, statuses

    def register_pending_keys(self, session_id, data_keys):
        for key in data_keys:
            session_data_key = (session_id, key)
            if session_data_key not in self._data_metas \
                    or self._data_metas[session_data_key].status == ReceiveStatus.ERROR:
                self._update_data_meta(session_id, key, status=ReceiveStatus.PENDING,
                                       callback_args=(), callback_kwargs={})

    def filter_receiving_keys(self, session_id, data_keys):
        keys = []
        receiving_status = (ReceiveStatus.PENDING, ReceiveStatus.RECEIVING)
        for k in data_keys:
            try:
                if self._data_metas[(session_id, k)].status in receiving_status:
                    keys.append(k)
            except KeyError:
                pass
        return keys

    def add_keys_callback(self, session_id, data_keys, callback):
        cb_id = self._max_callback_id
        self._max_callback_id += 1

        receiving_status = (ReceiveStatus.PENDING, ReceiveStatus.RECEIVING)
        registered_session_keys = []
        args, kwargs = (), {}
        for k in data_keys:
            session_data_key = (session_id, k)
            data_meta = self._data_metas[session_data_key]  # type: ReceiverDataMeta
            if data_meta.status in receiving_status:
                registered_session_keys.append(session_data_key)
                data_meta.callback_ids.append(cb_id)
            else:
                args, kwargs = data_meta.callback_args, data_meta.callback_kwargs

        if registered_session_keys:
            self._callback_id_to_callbacks[cb_id] = callback
            self._callback_id_to_keys[cb_id] = set(registered_session_keys)
            logger.debug('Callback for transferring %r registered', registered_session_keys)
        else:
            self._max_callback_id = cb_id
            self.tell_promise(callback, *args, **kwargs)

    def notify_keys_finish(self, session_id, data_keys, *args, **kwargs):
        keys_to_clear = []
        for data_key in data_keys:
            session_data_key = (session_id, data_key)
            try:
                data_meta = self._data_metas[session_data_key]  # type: ReceiverDataMeta
            except KeyError:
                logger.debug('Record of %s not found.', data_key)
                continue

            try:
                data_meta.callback_args = args
                data_meta.callback_kwargs = kwargs
                if kwargs.get('_accept', True):
                    data_meta.status = ReceiveStatus.RECEIVED
                else:
                    data_meta.status = ReceiveStatus.ERROR

                cb_ids = data_meta.callback_ids
                data_meta.callback_ids = []
                if not cb_ids:
                    continue

                kwargs['_wait'] = False
                notified = 0
                for cb_id in cb_ids:
                    cb_keys = self._callback_id_to_keys[cb_id]
                    cb_keys.remove(session_data_key)
                    if not cb_keys:
                        del self._callback_id_to_keys[cb_id]
                        cb = self._callback_id_to_callbacks.pop(cb_id)
                        notified += 1
                        self.tell_promise(cb, *args, **kwargs)
                logger.debug('%d transfer listeners of %s notified.', notified, data_key)
            finally:
                if data_meta.status == ReceiveStatus.RECEIVED:
                    keys_to_clear.append(data_key)
        self.ref().clear_keys(session_id, keys_to_clear, _tell=True, _delay=5)

    def clear_keys(self, session_id, keys):
        for k in keys:
            self._data_metas.pop((session_id, k), None)


class ReceiverWorkerActor(WorkerActor):
    """
    Actor handling receiving data from a SenderActor
    """
    def __init__(self):
        super().__init__()
        self._chunk_holder_ref = None
        self._dispatch_ref = None
        self._receiver_manager_ref = None
        self._events_ref = None
        self._status_ref = None

        self._data_writers = dict()
        self._writing_futures = dict()
        self._data_metas = dict()

    def post_create(self):
        from .events import EventsActor
        from .status import StatusActor
        from .dispatcher import DispatchActor

        super().post_create()

        self._events_ref = self.ctx.actor_ref(EventsActor.default_uid())
        if not self.ctx.has_actor(self._events_ref):
            self._events_ref = None

        self._status_ref = self.ctx.actor_ref(StatusActor.default_uid())
        if not self.ctx.has_actor(self._status_ref):
            self._status_ref = None

        self._receiver_manager_ref = self.ctx.actor_ref(ReceiverManagerActor.default_uid())
        if not self.ctx.has_actor(self._receiver_manager_ref):
            self._receiver_manager_ref = None

        self._dispatch_ref = self.promise_ref(DispatchActor.default_uid())
        self._dispatch_ref.register_free_slot(self.uid, 'receiver')

    @log_unhandled
    def check_status(self, session_id, chunk_key):
        """
        Check if data exists or is being transferred in the target worker
        :param session_id: session id
        :param chunk_key: chunk key
        """
        session_chunk_key = (session_id, chunk_key)

        if self.storage_client.get_data_locations(session_id, [chunk_key])[0]:
            return ReceiveStatus.RECEIVED
        if session_chunk_key in self._data_writers:
            # data still being transferred
            return ReceiveStatus.RECEIVING
        return ReceiveStatus.NOT_STARTED

    @promise.reject_on_exception
    @log_unhandled
    def create_data_writers(self, session_id, chunk_keys, data_sizes, sender_ref,
                            ensure_cached=True, timeout=0, pin_token=None,
                            use_promise=True, callback=None):
        """
        Create a data writer for subsequent data transfer. The writer can either work on
        shared storage or spill.

        :param session_id: session id
        :param chunk_keys: chunk keys
        :param data_sizes: uncompressed data sizes
        :param sender_ref: ActorRef of SenderActor
        :param ensure_cached: if True, the data should be stored in shared memory,
                              otherwise spill is acceptable
        :param timeout: timeout if the chunk receiver does not close
        :param pin_token: token to pin the data
        :param use_promise: if True, we use promise callback to notify accomplishment
                            of writer creation, otherwise the function returns directly
                            and when sill is needed, a StorageFull will be raised instead.
        :param callback: promise callback
        """
        promises = []
        failed = False
        if options.vineyard.enabled:
            device_order = [DataStorageDevice.VINEYARD]  # pragma: no cover
        else:
            device_order = [DataStorageDevice.SHARED_MEMORY]
        source_address = sender_ref.address if sender_ref is not None else None
        if not ensure_cached:
            device_order += [DataStorageDevice.DISK]

        def _handle_accept_key(key, writer):
            if failed:
                writer.close(finished=False)
            else:
                self._data_writers[(session_id, key)] = writer

        @log_unhandled
        def _handle_reject_key(key, *exc):
            nonlocal failed
            if self.check_status(session_id, key) == ReceiveStatus.RECEIVED:
                logger.debug('Chunk %s already received', key)
            else:
                logger.debug('Rejecting %s from putting into plasma.', key)
                failed = True
                self._stop_transfer_with_exc(session_id, chunk_keys, exc)
                if callback is not None:
                    self.tell_promise(callback, *exc, _accept=False)

        # configure timeout callback
        if timeout:
            self.ref().handle_receive_timeout(session_id, chunk_keys, _delay=timeout, _tell=True)

        for chunk_key, data_size in zip(chunk_keys, data_sizes):
            self._data_metas[(session_id, chunk_key)] = ReceiverDataMeta(
                start_time=time.time(), chunk_size=data_size, source_address=source_address)

            use_device_order = device_order \
                if not isinstance(chunk_key, tuple) or not options.worker.write_shuffle_to_disk \
                else [DataStorageDevice.DISK]
            if use_promise:
                promises.append(self.storage_client.create_writer(
                    session_id, chunk_key, data_size, use_device_order, packed=True,
                    pin_token=pin_token, _promise=True)
                    .then(functools.partial(_handle_accept_key, chunk_key),
                          functools.partial(_handle_reject_key, chunk_key)))
            else:
                try:
                    _writer = self.storage_client.create_writer(
                        session_id, chunk_key, data_size, use_device_order, packed=True,
                        pin_token=pin_token, _promise=False)
                    _handle_accept_key(chunk_key, _writer)
                    return self.address, None
                except:  # noqa: E722
                    _handle_reject_key(chunk_key, *sys.exc_info())
                    raise

        promise.all_(promises).then(lambda *_: self.tell_promise(callback))

    def _wait_unfinished_writing(self, session_id, data_key):
        try:
            self._writing_futures[(session_id, data_key)].result()
            del self._writing_futures[(session_id, data_key)]
        except KeyError:
            pass

    @log_unhandled
    def receive_data_part(self, session_id, chunk_keys, end_marks, *data_parts):
        """
        Receive data part from sender
        :param session_id: session id
        :param chunk_keys: chunk keys
        :param end_marks: array with same number of boolean elements as chunk keys.
                          if one element is True, the corresponding data in data_parts
                          is the last part of the chunk.
        :param data_parts: data parts to be written
        """
        try:
            finished_keys, finished_meta_keys, finished_metas = [], [], []
            for chunk_key, data_part, end_mark in zip(chunk_keys, data_parts, end_marks):
                self._wait_unfinished_writing(session_id, chunk_key)
                session_chunk_key = (session_id, chunk_key)
                try:
                    data_meta = self._data_metas[session_chunk_key]  # type: ReceiverDataMeta

                    # if error occurred, interrupts
                    if data_meta.status == ReceiveStatus.ERROR:
                        raise data_meta.callback_args[1].with_traceback(data_meta.callback_args[2])
                    writer = self._data_writers[session_chunk_key]
                    pool = writer.get_io_pool()
                    self._writing_futures[session_chunk_key] = pool.submit(
                        writer.write, data_part)

                    if end_mark:
                        finished_keys.append(chunk_key)
                        if not isinstance(chunk_key, tuple):
                            finished_meta_keys.append(chunk_key)
                            finished_metas.append(WorkerMeta(chunk_size=data_meta.chunk_size,
                                                             workers=(self.address,)))
                except:  # noqa: E722
                    self._stop_transfer_with_exc(session_id, chunk_keys, sys.exc_info())
                    raise

            if finished_keys:
                for chunk_key in finished_keys:
                    session_chunk_key = (session_id, chunk_key)
                    data_meta = self._data_metas[session_chunk_key]  # type: ReceiverDataMeta

                    self._wait_unfinished_writing(session_id, chunk_key)
                    # update transfer speed stats
                    if self._status_ref:
                        time_delta = time.time() - data_meta.start_time
                        self._status_ref.update_mean_stats(
                            'net_transfer_speed', data_meta.chunk_size * 1.0 / time_delta,
                            _tell=True, _wait=False)
                    self._data_writers[session_chunk_key].close()
                    data_meta.status = ReceiveStatus.RECEIVED
                    logger.debug('Transfer for data %s finished.', chunk_key)
                    del self._data_writers[session_chunk_key]

                self._invoke_finish_callbacks(session_id, finished_keys)
            if finished_meta_keys:
                self.get_meta_client().batch_set_chunk_meta(session_id, finished_meta_keys, finished_metas)
        finally:
            del data_parts

    def _is_receive_running(self, session_id, chunk_key):
        receive_done_statuses = (ReceiveStatus.ERROR, ReceiveStatus.RECEIVED)
        try:
            return self._data_metas[(session_id, chunk_key)].status not in receive_done_statuses
        except KeyError:
            return False

    @log_unhandled
    def cancel_receive(self, session_id, chunk_keys, exc_info=None):
        """
        Cancel data receive by returning an ExecutionInterrupted
        :param session_id: session id
        :param chunk_keys: chunk keys
        :param exc_info: exception to raise
        """
        receiving_keys = []
        for k in chunk_keys:
            if self._is_receive_running(session_id, k):
                receiving_keys.append(k)
            self._wait_unfinished_writing(session_id, k)

        logger.debug('Transfer for %r cancelled.', chunk_keys)

        if exc_info is None:
            exc_info = build_exc_info(ExecutionInterrupted)

        self._stop_transfer_with_exc(session_id, receiving_keys, exc_info)

    @log_unhandled
    def notify_dead_senders(self, dead_workers):
        """
        When some peer workers are dead, corresponding receivers will be cancelled
        :param dead_workers: endpoints of dead workers
        """
        dead_workers = set(dead_workers)
        exc_info = build_exc_info(WorkerDead)
        session_to_keys = defaultdict(set)
        for session_chunk_key in self._data_writers.keys():
            if self._data_metas[session_chunk_key].source_address in dead_workers:
                session_to_keys[session_chunk_key[0]].add(session_chunk_key[1])
        for session_id, data_keys in session_to_keys.items():
            self.ref().cancel_receive(session_id, list(data_keys), exc_info=exc_info, _tell=True)

    @log_unhandled
    def handle_receive_timeout(self, session_id, chunk_keys):
        if not any(self._is_receive_running(session_id, k) for k in chunk_keys):
            # if transfer already finishes, no needs to report timeout
            return
        logger.debug('Transfer for %r timed out, cancelling.', chunk_keys)
        self._stop_transfer_with_exc(session_id, chunk_keys, build_exc_info(TimeoutError))

    def _stop_transfer_with_exc(self, session_id, chunk_keys, exc):
        for chunk_key in chunk_keys:
            self._wait_unfinished_writing(session_id, chunk_key)

        if not isinstance(exc[1], ExecutionInterrupted):
            logger.exception('Error occurred in receiving %r. Cancelling transfer.',
                             chunk_keys, exc_info=exc)

        for chunk_key in chunk_keys:
            session_chunk_key = (session_id, chunk_key)

            # stop and close data writer
            try:
                # transfer is not finished yet, we need to clean up unfinished stuffs
                self._data_writers[session_chunk_key].close(finished=False)
                del self._data_writers[session_chunk_key]
            except KeyError:
                # transfer finished and writer cleaned, no need to clean up
                pass

            try:
                data_meta = self._data_metas[session_chunk_key]  # type: ReceiverDataMeta
                data_meta.status = ReceiveStatus.ERROR
            except KeyError:
                pass

        self._invoke_finish_callbacks(session_id, chunk_keys, *exc, **dict(_accept=False))

    def _invoke_finish_callbacks(self, session_id, chunk_keys, *args, **kwargs):
        # invoke registered callbacks for chunk
        for k in chunk_keys:
            try:
                data_meta = self._data_metas.pop((session_id, k))  # type: ReceiverDataMeta
            except KeyError:
                continue

            if data_meta.transfer_event_id is not None and self._events_ref is not None:
                self._events_ref.close_event(data_meta.transfer_event_id, _tell=True, _wait=False)
            if not kwargs.get('_accept', True):
                if not data_meta.callback_args or data_meta.callback_args[0] is ExecutionInterrupted:
                    data_meta.callback_args = args
                    data_meta.callback_kwargs = kwargs
                else:
                    args = data_meta.callback_args
                    kwargs = data_meta.callback_kwargs
            else:
                data_meta.callback_args = args
                data_meta.callback_kwargs = kwargs
        kwargs['_tell'] = True
        if self._receiver_manager_ref:
            self._receiver_manager_ref.notify_keys_finish(session_id, chunk_keys, *args, **kwargs)


class ResultCopyActor(WorkerActor):
    def start_copy(self, session_id, chunk_key, targets):
        locations = [v[1] for v in self.storage_client.get_data_locations(session_id, [chunk_key])[0]]
        if set(locations).intersection(targets):
            return
        ev = self.ctx.event()
        self.storage_client.copy_to(session_id, [chunk_key], targets) \
            .then(lambda *_: ev.set())
        return ev


class ResultSenderActor(WorkerActor):
    """
    Actor handling sending result to user client
    """
    def __init__(self):
        super().__init__()
        self._result_copy_ref = None
        self._serialize_pool = None

    def post_create(self):
        super().post_create()
        self._serialize_pool = self.ctx.threadpool(1)
        self._result_copy_ref = self.ctx.create_actor(ResultCopyActor, uid=ResultCopyActor.default_uid())

    def pre_destroy(self):
        self._result_copy_ref.destroy()
        super().pre_destroy()

    def fetch_batch_data(self, session_id, chunk_keys, index_objs=None, compression_type=None):
        results = []
        if index_objs is not None:
            for chunk_key, index_obj in zip(chunk_keys, index_objs):
                results.append(self.fetch_data(session_id, chunk_key, index_obj, compression_type=compression_type))
        else:
            for chunk_key in chunk_keys:
                results.append(self.fetch_data(session_id, chunk_key, compression_type=compression_type))
        return results

    def fetch_data(self, session_id, chunk_key, index_obj=None, compression_type=None):
        logger.debug('Sending data %s from %s, using compression: %r', chunk_key, self.uid, compression_type)
        if options.vineyard.enabled:
            compression_type = None
        else:
            compression_type = compression_type or dataserializer.CompressType(options.worker.transfer_compression)

        if options.vineyard.enabled:
            in_memory_device = DataStorageDevice.VINEYARD
        else:
            in_memory_device = DataStorageDevice.SHARED_MEMORY

        if index_obj is None:
            target_devs = [in_memory_device, DataStorageDevice.DISK]  # pragma: no cover
            ev = self._result_copy_ref.start_copy(session_id, chunk_key, target_devs)
            if ev:
                ev.wait(options.worker.prepare_data_timeout)

            reader = self.storage_client.create_reader(
                session_id, chunk_key, target_devs, packed=True,
                packed_compression=compression_type, _promise=False)

            with reader:
                pool = reader.get_io_pool()
                return pool.submit(reader.read).result()
        else:
            try:
                value = self.storage_client.get_object(
                    session_id, chunk_key, [in_memory_device], _promise=False)
            except IOError:
                reader = self.storage_client.create_reader(
                    session_id, chunk_key, [DataStorageDevice.DISK], packed=False, _promise=False)
                with reader:
                    pool = reader.get_io_pool()
                    value = dataserializer.deserialize(pool.submit(reader.read).result())

            try:
                sliced_value = value.iloc[tuple(index_obj)]
            except AttributeError:
                sliced_value = value[tuple(index_obj)]

            return self._serialize_pool.submit(
                dataserializer.dumps, sliced_value, compress=compression_type).result()


def put_remote_chunk(session_id, chunk_key, data, receiver_manager_ref):
    """
    Put a chunk to target machine using given receiver_ref
    """
    from .dataio import ArrowBufferIO
    buf = dataserializer.serialize(data).to_buffer()
    receiver_ref, _ = receiver_manager_ref.create_data_writers(
        session_id, [chunk_key], [buf.size], None, ensure_cached=False, use_promise=False)
    receiver_ref = receiver_manager_ref.ctx.actor_ref(receiver_ref)
    block_size = options.worker.transfer_block_size

    reader = None
    try:
        reader = ArrowBufferIO(buf, 'r', block_size=block_size)
        futures = []
        while True:
            next_part = reader.read(block_size)
            is_last = not next_part or len(next_part) < block_size
            [f.result() for f in futures]
            futures.append(receiver_ref.receive_data_part(
                session_id, [chunk_key], [is_last], next_part, _wait=False))
            if is_last:
                [f.result() for f in futures]
                break
    except:  # noqa: E722
        receiver_ref.cancel_receive(session_id, [chunk_key])
        raise
    finally:
        if reader:
            reader.close()
        del reader



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
import operator

import numpy as np

from ... import opcodes as OperandDef
from ... import tensor as mt
from ...serialize import AnyField, TupleField, KeyField, BoolField
from ...tiles import TilesError
from ...context import get_context
from ...utils import check_chunks_unknown_shape, recursive_tile
from ..core import TENSOR_TYPE, TENSOR_CHUNK_TYPE, TensorOrder
from ..operands import TensorOperand, TensorOperandMixin
from ..datasource import tensor as astensor
from ..arithmetic.utils import chunk_tree_add
from ..utils import is_asc_sorted
from ..array_utils import as_same_device, device

# note: some logic of this file were adopted from `numpy/lib/histograms`


def _ptp(range_):
    """Peak-to-peak value of x.

    This implementation avoids the problem of signed integer arrays having a
    peak-to-peak value that cannot be represented with the array's data type.
    This function returns an unsigned value for signed integer arrays.
    """
    return _unsigned_subtract(*range_[::-1])


class HistBinSelector:
    def __init__(self, histogram_bin_edges_op, x, range, raw_range):
        self._op = histogram_bin_edges_op
        self._x = x
        self._range = range
        self._raw_range = raw_range

    def check(self):
        if len(self._op._calc_bin_edges_dependencies) == 0:
            # not checked before
            width = self()
            if width is None:
                return
            err = TilesError('bin edges calculation requires '
                             'some dependencies executed first')
            self._op._calc_bin_edges_dependencies = [width]
            recursive_tile(width)
            err.partial_tiled_chunks = [c.data for c in width.chunks]
            raise err

    def __call__(self):
        return

    def get_result(self):
        ctx = get_context()
        width = ctx.get_chunk_results(
            [self._op._calc_bin_edges_dependencies[0].chunks[0].key])[0]
        return width


class HistBinSqrtSelector(HistBinSelector):
    """
    Square root histogram bin estimator.

    Bin width is inversely proportional to the data size. Used by many
    programs for its simplicity.
    """

    def get_result(self):
        return _ptp(self._raw_range) / np.sqrt(self._x.size)


class HistBinSturgesSelector(HistBinSelector):
    """
    Sturges histogram bin estimator.

    A very simplistic estimator based on the assumption of normality of
    the data. This estimator has poor performance for non-normal data,
    which becomes especially obvious for large data sets. The estimate
    depends only on size of the data.
    """

    def get_result(self):
        return _ptp(self._raw_range) / (np.log2(self._x.size) + 1.0)


class HistBinRiceSelector(HistBinSelector):
    """
    Rice histogram bin estimator.

    Another simple estimator with no normality assumption. It has better
    performance for large data than Sturges, but tends to overestimate
    the number of bins. The number of bins is proportional to the cube
    root of data size (asymptotically optimal). The estimate depends
    only on size of the data.
    """

    def get_result(self):
        return _ptp(self._raw_range) / (2.0 * self._x.size ** (1.0 / 3))


class HistBinScottSelector(HistBinSelector):
    """
    Scott histogram bin estimator.

    The binwidth is proportional to the standard deviation of the data
    and inversely proportional to the cube root of data size
    (asymptotically optimal).
    """

    def __call__(self):
        return (24.0 * np.pi**0.5 / self._x.size)**(1.0 / 3.0) * mt.std(self._x)


class HistBinStoneSelector(HistBinSelector):
    """
    Histogram bin estimator based on minimizing the estimated integrated squared error (ISE).

    The number of bins is chosen by minimizing the estimated ISE against the unknown true distribution.
    The ISE is estimated using cross-validation and can be regarded as a generalization of Scott's rule.
    https://en.wikipedia.org/wiki/Histogram#Scott.27s_normal_reference_rule

    This paper by Stone appears to be the origination of this rule.
    http://digitalassets.lib.berkeley.edu/sdtr/ucb/text/34.pdf
    """

    def __call__(self):
        n = self._x.size
        ptp_x = _ptp(self._raw_range)

        if n <= 1 or ptp_x == 0:
            return

        nbins_upper_bound = max(100, int(np.sqrt(n)))
        candidates = []
        for nbins in range(1, nbins_upper_bound + 1):
            hh = ptp_x / nbins
            p_k = histogram(self._x, bins=nbins, range=self._range)[0] / n
            candidate = (2 - (n + 1) * p_k.dot(p_k)) / hh
            candidates.append(candidate)
        nbins = mt.stack(candidates).argmin() + 1
        return ptp_x / nbins

    def get_result(self):
        ptp_x = _ptp(self._raw_range)
        if self._x.size <= 1 or ptp_x == 0:
            return 0.0
        else:
            return super().get_result()


class HistBinDoaneSelector(HistBinSelector):
    """
    Doane's histogram bin estimator.

    Improved version of Sturges' formula which works better for
    non-normal data. See
    stats.stackexchange.com/questions/55134/doanes-formula-for-histogram-binning
    """

    def __call__(self):
        x = self._x
        if x.size <= 2:
            return

        sg1 = np.sqrt(6.0 * (x.size - 2) / ((x.size + 1.0) * (x.size + 3)))
        sigma = mt.std(x)
        g1 = mt.mean(((x - mt.mean(x)) / sigma)**3)
        ret = _ptp(self._raw_range) / (1.0 + np.log2(x.size) +
                                      mt.log2(1.0 + mt.absolute(g1) / sg1))
        return mt.where(sigma > 0.0, ret, 0.0)

    def get_result(self):
        if self._x.size <= 2:
            return 0.0
        else:
            return super().get_result()


class HistBinFdSelector(HistBinSelector):
    """
    The Freedman-Diaconis histogram bin estimator.

    The Freedman-Diaconis rule uses interquartile range (IQR) to
    estimate binwidth. It is considered a variation of the Scott rule
    with more robustness as the IQR is less affected by outliers than
    the standard deviation. However, the IQR depends on fewer points
    than the standard deviation, so it is less accurate, especially for
    long tailed distributions.

    If the IQR is 0, this function returns 1 for the number of bins.
    Binwidth is inversely proportional to the cube root of data size
    (asymptotically optimal).
    """

    def __call__(self):
        iqr = mt.subtract(*mt.percentile(self._x, [75, 25]))
        return 2.0 * iqr * self._x.size ** (-1.0 / 3.0)


class HistBinAutoSelector(HistBinSelector):
    """
    Histogram bin estimator that uses the minimum width of the
    Freedman-Diaconis and Sturges estimators if the FD bandwidth is non zero
    and the Sturges estimator if the FD bandwidth is 0.

    The FD estimator is usually the most robust method, but its width
    estimate tends to be too large for small `x` and bad for data with limited
    variance. The Sturges estimator is quite good for small (<1000) datasets
    and is the default in the R language. This method gives good off the shelf
    behaviour.

    If there is limited variance the IQR can be 0, which results in the
    FD bin width being 0 too. This is not a valid bin width, so
    ``np.histogram_bin_edges`` chooses 1 bin instead, which may not be optimal.
    If the IQR is 0, it's unlikely any variance based estimators will be of
    use, so we revert to the sturges estimator, which only uses the size of the
    dataset in its calculation.
    """

    def __init__(self, histogram_bin_edges_op, x, range, raw_range):
        super().__init__(histogram_bin_edges_op, x, range, raw_range)
        self._bin_fd = HistBinFdSelector(
            histogram_bin_edges_op, x, range, raw_range)
        self._bin_sturges = HistBinSturgesSelector(
            histogram_bin_edges_op, x, range, raw_range)

    def __call__(self):
        return self._bin_fd()

    def get_result(self):
        fd_bw = super().get_result()
        sturges_bw = self._bin_sturges.get_result()
        if fd_bw:
            return min(fd_bw, sturges_bw)
        else:
            # limited variance, so we return a len dependent bw estimator
            return sturges_bw


# Private dict initialized at module load time
_hist_bin_selectors = {'stone': HistBinStoneSelector,
                       'auto': HistBinAutoSelector,
                       'doane': HistBinDoaneSelector,
                       'fd': HistBinFdSelector,
                       'rice': HistBinRiceSelector,
                       'scott': HistBinScottSelector,
                       'sqrt': HistBinSqrtSelector,
                       'sturges': HistBinSturgesSelector}


def _ravel_and_check_weights(a, weights):
    """ Check a and weights have matching shapes, and ravel both """
    a = astensor(a)

    # Ensure that the array is a "subtractable" dtype
    if a.dtype == np.bool_:
        warnings.warn(f"Converting input from {a.dtype} to {np.uint8} for compatibility.",
                      RuntimeWarning, stacklevel=3)
        a = a.astype(np.uint8)

    if weights is not None:
        weights = astensor(weights)
        if weights.shape != a.shape:
            raise ValueError(
                'weights should have the same shape as a.')
        weights = weights.ravel()
    a = a.ravel()
    return a, weights


def _check_range(range):
    first_edge, last_edge = range
    if first_edge > last_edge:
        raise ValueError(
            'max must be larger than min in range parameter.')
    if not (np.isfinite(first_edge) and np.isfinite(last_edge)):
        raise ValueError(
            f"supplied range of [{first_edge}, {last_edge}] is not finite")
    return first_edge, last_edge


def _get_outer_edges(a, range):
    """
    Determine the outer bin edges to use, from either the data or the range
    argument
    """
    if range is not None:
        first_edge, last_edge = _check_range(range)
    else:
        assert a.size == 0
        # handle empty arrays. Can't determine range, so use 0-1.
        first_edge, last_edge = 0, 1

    # expand empty range to avoid divide by zero
    if first_edge == last_edge:
        first_edge = first_edge - 0.5
        last_edge = last_edge + 0.5

    return first_edge, last_edge


def _unsigned_subtract(a, b):
    """
    Subtract two values where a >= b, and produce an unsigned result

    This is needed when finding the difference between the upper and lower
    bound of an int16 histogram
    """
    # coerce to a single type
    signed_to_unsigned = {
        np.byte: np.ubyte,
        np.short: np.ushort,
        np.intc: np.uintc,
        np.int_: np.uint,
        np.longlong: np.ulonglong
    }
    dt = np.result_type(a, b)
    try:
        dt = signed_to_unsigned[dt.type]
    except KeyError:  # pragma: no cover
        return np.subtract(a, b, dtype=dt)
    else:
        # we know the inputs are integers, and we are deliberately casting
        # signed to unsigned
        return np.subtract(a, b, casting='unsafe', dtype=dt)


def _get_bin_edges(op, a, bins, range, weights):
    # parse the overloaded bins argument
    n_equal_bins = None
    bin_edges = None
    first_edge = None
    last_edge = None

    if isinstance(bins, str):
        # when `bins` is str, x.min() and x.max()
        # will be calculated in advance
        bin_name = bins
        if a.size > 0:
            assert range is not None

        raw_range = range
        first_edge, last_edge = _get_outer_edges(a, range)

        if a.size == 0:
            n_equal_bins = 1
        else:
            # Do not call selectors on empty arrays
            selector = _hist_bin_selectors[bin_name](op, a, (first_edge, last_edge), raw_range)
            selector.check()
            width = selector.get_result()
            if width:
                n_equal_bins = int(np.ceil(_unsigned_subtract(last_edge, first_edge) / width))
            else:
                # Width can be zero for some estimators, e.g. FD when
                # the IQR of the data is zero.
                n_equal_bins = 1

    elif mt.ndim(bins) == 0:
        first_edge, last_edge = _get_outer_edges(a, range)
        n_equal_bins = bins

    else:
        # cannot be Tensor, must be calculated first
        assert mt.ndim(bins) == 1 and not isinstance(bins, TENSOR_TYPE)
        bin_edges = np.asarray(bins)
        if not is_asc_sorted(bin_edges):
            raise ValueError(
                '`bins` must increase monotonically, when an array')

    if n_equal_bins is not None:
        # numpy gh-10322 means that type resolution rules are dependent on array
        # shapes. To avoid this causing problems, we pick a type now and stick
        # with it throughout.
        bin_type = np.result_type(first_edge, last_edge, a)
        if np.issubdtype(bin_type, np.integer):
            bin_type = np.result_type(bin_type, float)

        # bin edges must be computed
        bin_edges = mt.linspace(
            first_edge, last_edge, n_equal_bins + 1,
            endpoint=True, dtype=bin_type, gpu=op.gpu)
        return bin_edges, (first_edge, last_edge, n_equal_bins)
    else:
        return mt.tensor(bin_edges), None


class TensorHistogramBinEdges(TensorOperand, TensorOperandMixin):
    __slots__ = '_calc_bin_edges_dependencies',
    _op_type_ = OperandDef.HISTOGRAM_BIN_EDGES

    _input = KeyField('input')
    _bins = AnyField('bins')
    _range = TupleField('range')
    _weights = KeyField('weights')
    _input_min = KeyField('input_min')
    _input_max = KeyField('input_max')
    _uniform_bins = TupleField('uniform_bins')

    def __init__(self, input=None, bins=None, range=None, weights=None,
                 input_min=None, input_max=None, dtype=None, **kw):
        super().__init__(_input=input, _bins=bins, _range=range, _weights=weights,
                         _input_min=input_min, _input_max=input_max, _dtype=dtype, **kw)
        if getattr(self, '_calc_bin_edges_dependencies', None) is None:
            self._calc_bin_edges_dependencies = []

    @property
    def input(self):
        return self._input

    @property
    def bins(self):
        return self._bins

    @property
    def range(self):
        return self._range

    @property
    def weights(self):
        return self._weights

    @property
    def input_min(self):
        return self._input_min

    @property
    def input_max(self):
        return self._input_max

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        inputs_iter = iter(self._inputs)
        self._input = next(inputs_iter)
        if isinstance(self._bins, (TENSOR_TYPE, TENSOR_CHUNK_TYPE)):
            self._bins = next(inputs_iter)
        if self._weights is not None:
            self._weights = next(inputs_iter)
        if self._input_min is not None:
            self._input_min = next(inputs_iter)
        if self._input_max is not None:
            self._input_max = next(inputs_iter)

    def __call__(self, a, bins, range, weights):
        if range is not None:
            _check_range(range)
        if isinstance(bins, str):
            # string, 'auto', 'stone', ...
            # shape is unknown
            bin_name = bins
            # if `bins` is a string for an automatic method,
            # this will replace it with the number of bins calculated
            if bin_name not in _hist_bin_selectors:
                raise ValueError(
                    f"{bin_name!r} is not a valid estimator for `bins`")
            if weights is not None:
                raise TypeError("Automated estimation of the number of "
                                "bins is not supported for weighted data")
            if isinstance(range, tuple) and len(range) == 2:
                # if `bins` is a string, e.g. 'auto', 'stone'...,
                # and `range` provided as well,
                # `a` should be trimmed first
                first_edge, last_edge = _get_outer_edges(a, range)
                a = a[(a >= first_edge) & (a <= last_edge)]
            shape = (np.nan,)
        elif mt.ndim(bins) == 0:
            try:
                n_equal_bins = operator.index(bins)
            except TypeError:  # pragma: no cover
                raise TypeError(
                    '`bins` must be an integer, a string, or an array')
            if n_equal_bins < 1:
                raise ValueError('`bins` must be positive, when an integer')
            shape = (bins + 1,)
        elif mt.ndim(bins) == 1:
            if not isinstance(bins, TENSOR_TYPE):
                bins = np.asarray(bins)
                if not is_asc_sorted(bins):
                    raise ValueError(
                        '`bins` must increase monotonically, when an array')
            shape = astensor(bins).shape
        else:
            raise ValueError('`bins` must be 1d, when an array')

        inputs = [a]
        if isinstance(bins, TENSOR_TYPE):
            inputs.append(bins)
        if weights is not None:
            inputs.append(weights)
        if (a.size > 0 or np.isnan(a.size)) and \
                (isinstance(bins, str) or mt.ndim(bins) == 0) and not range:
            # for bins that is str or integer, requires min max calculated first
            # dims need to be kept in case a is empty which causes errors in reduction
            input_min = self._input_min = a.min(keepdims=True)
            inputs.append(input_min)
            input_max = self._input_max = a.max(keepdims=True)
            inputs.append(input_max)

        return self.new_tensor(inputs, shape=shape, order=TensorOrder.C_ORDER)

    @classmethod
    def tile(cls, op):
        ctx = get_context()
        range_ = op.range
        if isinstance(op.bins, str):
            check_chunks_unknown_shape([op.input], TilesError)
        if op.input_min is not None:
            # check if input min and max are calculated
            min_max_chunk_keys = \
                [inp.chunks[0].key for inp in (op.input_min, op.input_max)]
            metas = ctx.get_chunk_metas(min_max_chunk_keys)
            if any(meta is None for meta in metas):
                raise TilesError('`input_min` or `input_max` need be executed first')
            range_results = ctx.get_chunk_results(min_max_chunk_keys)
            # make sure returned bounds are valid
            if all(x.size > 0 for x in range_results):
                range_ = tuple(x[0] for x in range_results)
        if isinstance(op.bins, TENSOR_TYPE):
            # `bins` is a Tensor, needs to be calculated first
            bins_chunk_keys = [c.key for c in op.bins.chunks]
            metas = ctx.get_chunk_metas(bins_chunk_keys)
            if any(meta is None for meta in metas):
                raise TilesError('`bins` should be executed first if it\'s a tensor')
            bin_datas = ctx.get_chunk_results(bins_chunk_keys)
            bins = np.concatenate(bin_datas)
        else:
            bins = op.bins

        bin_edges, _ = _get_bin_edges(op, op.input, bins, range_, op.weights)
        bin_edges = bin_edges._inplace_tile()
        return [bin_edges]


def histogram_bin_edges(a, bins=10, range=None, weights=None):
    r"""
    Function to calculate only the edges of the bins used by the `histogram`
    function.

    Parameters
    ----------
    a : array_like
        Input data. The histogram is computed over the flattened tensor.
    bins : int or sequence of scalars or str, optional
        If `bins` is an int, it defines the number of equal-width
        bins in the given range (10, by default). If `bins` is a
        sequence, it defines the bin edges, including the rightmost
        edge, allowing for non-uniform bin widths.

        If `bins` is a string from the list below, `histogram_bin_edges` will use
        the method chosen to calculate the optimal bin width and
        consequently the number of bins (see `Notes` for more detail on
        the estimators) from the data that falls within the requested
        range. While the bin width will be optimal for the actual data
        in the range, the number of bins will be computed to fill the
        entire range, including the empty portions. For visualisation,
        using the 'auto' option is suggested. Weighted data is not
        supported for automated bin size selection.

        'auto'
            Maximum of the 'sturges' and 'fd' estimators. Provides good
            all around performance.

        'fd' (Freedman Diaconis Estimator)
            Robust (resilient to outliers) estimator that takes into
            account data variability and data size.

        'doane'
            An improved version of Sturges' estimator that works better
            with non-normal datasets.

        'scott'
            Less robust estimator that that takes into account data
            variability and data size.

        'stone'
            Estimator based on leave-one-out cross-validation estimate of
            the integrated squared error. Can be regarded as a generalization
            of Scott's rule.

        'rice'
            Estimator does not take variability into account, only data
            size. Commonly overestimates number of bins required.

        'sturges'
            R's default method, only accounts for data size. Only
            optimal for gaussian data and underestimates number of bins
            for large non-gaussian datasets.

        'sqrt'
            Square root (of data size) estimator, used by Excel and
            other programs for its speed and simplicity.

    range : (float, float), optional
        The lower and upper range of the bins.  If not provided, range
        is simply ``(a.min(), a.max())``.  Values outside the range are
        ignored. The first element of the range must be less than or
        equal to the second. `range` affects the automatic bin
        computation as well. While bin width is computed to be optimal
        based on the actual data within `range`, the bin count will fill
        the entire range including portions containing no data.

    weights : array_like, optional
        A tensor of weights, of the same shape as `a`.  Each value in
        `a` only contributes its associated weight towards the bin count
        (instead of 1). This is currently not used by any of the bin estimators,
        but may be in the future.

    Returns
    -------
    bin_edges : tensor of dtype float
        The edges to pass into `histogram`

    See Also
    --------
    histogram

    Notes
    -----
    The methods to estimate the optimal number of bins are well founded
    in literature, and are inspired by the choices R provides for
    histogram visualisation. Note that having the number of bins
    proportional to :math:`n^{1/3}` is asymptotically optimal, which is
    why it appears in most estimators. These are simply plug-in methods
    that give good starting points for number of bins. In the equations
    below, :math:`h` is the binwidth and :math:`n_h` is the number of
    bins. All estimators that compute bin counts are recast to bin width
    using the `ptp` of the data. The final bin count is obtained from
    ``np.round(np.ceil(range / h))``.

    'auto' (maximum of the 'sturges' and 'fd' estimators)
        A compromise to get a good value. For small datasets the Sturges
        value will usually be chosen, while larger datasets will usually
        default to FD.  Avoids the overly conservative behaviour of FD
        and Sturges for small and large datasets respectively.
        Switchover point is usually :math:`a.size \approx 1000`.

    'fd' (Freedman Diaconis Estimator)
        .. math:: h = 2 \frac{IQR}{n^{1/3}}

        The binwidth is proportional to the interquartile range (IQR)
        and inversely proportional to cube root of a.size. Can be too
        conservative for small datasets, but is quite good for large
        datasets. The IQR is very robust to outliers.

    'scott'
        .. math:: h = \sigma \sqrt[3]{\frac{24 * \sqrt{\pi}}{n}}

        The binwidth is proportional to the standard deviation of the
        data and inversely proportional to cube root of ``x.size``. Can
        be too conservative for small datasets, but is quite good for
        large datasets. The standard deviation is not very robust to
        outliers. Values are very similar to the Freedman-Diaconis
        estimator in the absence of outliers.

    'rice'
        .. math:: n_h = 2n^{1/3}

        The number of bins is only proportional to cube root of
        ``a.size``. It tends to overestimate the number of bins and it
        does not take into account data variability.

    'sturges'
        .. math:: n_h = \log _{2}n+1

        The number of bins is the base 2 log of ``a.size``.  This
        estimator assumes normality of data and is too conservative for
        larger, non-normal datasets. This is the default method in R's
        ``hist`` method.

    'doane'
        .. math:: n_h = 1 + \log_{2}(n) +
                        \log_{2}(1 + \frac{|g_1|}{\sigma_{g_1}})

            g_1 = mean[(\frac{x - \mu}{\sigma})^3]

            \sigma_{g_1} = \sqrt{\frac{6(n - 2)}{(n + 1)(n + 3)}}

        An improved version of Sturges' formula that produces better
        estimates for non-normal datasets. This estimator attempts to
        account for the skew of the data.

    'sqrt'
        .. math:: n_h = \sqrt n

        The simplest and fastest estimator. Only takes into account the
        data size.

    Examples
    --------
    >>> import mars.tensor as mt
    >>> arr = mt.array([0, 0, 0, 1, 2, 3, 3, 4, 5])
    >>> mt.histogram_bin_edges(arr, bins='auto', range=(0, 1)).execute()
    array([0.  , 0.25, 0.5 , 0.75, 1.  ])
    >>> mt.histogram_bin_edges(arr, bins=2).execute()
    array([0. , 2.5, 5. ])

    For consistency with histogram, a tensor of pre-computed bins is
    passed through unmodified:

    >>> mt.histogram_bin_edges(arr, [1, 2]).execute()
    array([1, 2])

    This function allows one set of bins to be computed, and reused across
    multiple histograms:

    >>> shared_bins = mt.histogram_bin_edges(arr, bins='auto')
    >>> shared_bins.execute()
    array([0., 1., 2., 3., 4., 5.])

    >>> group_id = mt.array([0, 1, 1, 0, 1, 1, 0, 1, 1])
    >>> a = arr[group_id == 0]
    >>> a.execute()
    array([0, 1, 3])
    >>> hist_0, _ = mt.histogram(a, bins=shared_bins).execute()
    >>> b = arr[group_id == 1]
    >>> b.execute()
    array([0, 0, 2, 3, 4, 5])
    >>> hist_1, _ = mt.histogram(b, bins=shared_bins).execute()

    >>> hist_0; hist_1
    array([1, 1, 0, 1, 0])
    array([2, 0, 1, 1, 2])

    Which gives more easily comparable results than using separate bins for
    each histogram:

    >>> hist_0, bins_0 = mt.histogram(a, bins='auto').execute()
    >>> hist_1, bins_1 = mt.histogram(b, bins='auto').execute()
    >>> hist_0; hist_1
    array([1, 1, 1])
    array([2, 1, 1, 2])
    >>> bins_0; bins_1
    array([0., 1., 2., 3.])
    array([0.  , 1.25, 2.5 , 3.75, 5.  ])

    """
    a, weights = _ravel_and_check_weights(a, weights)
    op = TensorHistogramBinEdges(input=a, bins=bins,
                                 range=range, weights=weights,
                                 dtype=a.dtype)
    return op(a, bins, range, weights)


class TensorHistogram(TensorOperand, TensorOperandMixin):
    _op_type_ = OperandDef.HISTOGRAM

    _input = KeyField('input')
    _bins = AnyField('bins')
    _range = TupleField('range')
    _weights = KeyField('weights')
    _density = BoolField('density')
    _ret_bins = BoolField('ret_bins')

    def __init__(self, input=None, bins=None, range=None, weights=None,
                 density=None, ret_bins=None, **kw):
        super().__init__(_input=input, _bins=bins, _range=range, _weights=weights,
                         _density=density, _ret_bins=ret_bins, **kw)

    @property
    def input(self):
        return self._input

    @property
    def bins(self):
        return self._bins

    @property
    def range(self):
        return self._range

    @property
    def weights(self):
        return self._weights

    @property
    def density(self):
        return self._density

    @property
    def ret_bins(self):
        return self._ret_bins

    @property
    def output_limit(self):
        return 1 if not self._ret_bins else 2

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        inputs_iter = iter(self._inputs)
        self._input = next(inputs_iter)
        if isinstance(self._bins, (TENSOR_TYPE, TENSOR_CHUNK_TYPE)):
            self._bins = next(inputs_iter)
        if self._weights is not None:
            self._weights = next(inputs_iter)

    def __call__(self, a, bins, range, weights):
        a, weights = _ravel_and_check_weights(a, weights)
        histogram_bin_edges_op = TensorHistogramBinEdges(
            input=a, bins=bins, range=range, weights=weights,
            dtype=np.dtype(np.float64))
        bins = self._bins = histogram_bin_edges_op(a, bins, range, weights)

        inputs = [histogram_bin_edges_op.input]
        if isinstance(bins, TENSOR_TYPE):
            inputs.append(bins)
        # Histogram is an integer or a float array depending on the weights.
        if weights is None:
            dtype = np.dtype(np.intp)
        else:
            inputs.append(weights)
            dtype = weights.dtype
        self._dtype = dtype

        hist = self.new_tensor(inputs, shape=(bins.size - 1,),
                               order=TensorOrder.C_ORDER)
        return mt.ExecutableTuple([hist, bins])

    @classmethod
    def tile(cls, op):
        bins = op.bins.rechunk(op.bins.shape)
        shape = (bins.size - 1,)
        out = op.outputs[0]
        weights = None
        if op.weights is not None:
            # make input and weights have the same nsplits
            weights = op.weights.rechunk(op.input.nsplits)._inplace_tile()

        out_chunks = []
        for chunk in op.input.chunks:
            chunk_op = op.copy().reset_key()
            chunk_op._range = None
            chunk_op._ret_bins = False
            chunk_op._density = False
            chunk_inputs = [chunk, bins.chunks[0]]
            if weights is not None:
                weights_chunk = weights.cix[chunk.index]
                chunk_inputs.append(weights_chunk)
            out_chunk = chunk_op.new_chunk(chunk_inputs, shape=shape,
                                           index=chunk.index, order=out.order)
            out_chunks.append(out_chunk)

        # merge chunks together
        chunk = chunk_tree_add(out.dtype, out_chunks, (0,), shape)
        new_op = op.copy()
        n = new_op.new_tensor(op.inputs, shape=shape, order=out.order,
                              chunks=[chunk], nsplits=((shape[0],),))
        if op.density:
            db = mt.array(mt.diff(bins), float)
            hist = n / db / n.sum()
            recursive_tile(hist)
            return [hist]
        else:
            return [n]

    @classmethod
    def execute(cls, ctx, op):
        inputs, device_id, xp = as_same_device(
            [ctx[inp.key] for inp in op.inputs], device=op.device, ret_extra=True)
        a = inputs[0]
        bins = inputs[1] if isinstance(op.bins, TENSOR_CHUNK_TYPE) else op.bins
        weights = None
        if op.weights is not None:
            weights = inputs[-1]
        with device(device_id):
            hist, bin_edges = xp.histogram(a, bins=bins, range=op.range,
                                           weights=weights, density=op.density)
            ctx[op.outputs[0].key] = hist
            if op.ret_bins:
                ctx[op.outputs[1].key] = bin_edges


def histogram(a, bins=10, range=None, weights=None, density=None):
    r"""
    Compute the histogram of a set of data.

    Parameters
    ----------
    a : array_like
        Input data. The histogram is computed over the flattened tensor.
    bins : int or sequence of scalars or str, optional
        If `bins` is an int, it defines the number of equal-width
        bins in the given range (10, by default). If `bins` is a
        sequence, it defines a monotonically increasing tensor of bin edges,
        including the rightmost edge, allowing for non-uniform bin widths.

        If `bins` is a string, it defines the method used to calculate the
        optimal bin width, as defined by `histogram_bin_edges`.

    range : (float, float), optional
        The lower and upper range of the bins.  If not provided, range
        is simply ``(a.min(), a.max())``.  Values outside the range are
        ignored. The first element of the range must be less than or
        equal to the second. `range` affects the automatic bin
        computation as well. While bin width is computed to be optimal
        based on the actual data within `range`, the bin count will fill
        the entire range including portions containing no data.

    weights : array_like, optional
        A tensor of weights, of the same shape as `a`.  Each value in
        `a` only contributes its associated weight towards the bin count
        (instead of 1). If `density` is True, the weights are
        normalized, so that the integral of the density over the range
        remains 1.
    density : bool, optional
        If ``False``, the result will contain the number of samples in
        each bin. If ``True``, the result is the value of the
        probability *density* function at the bin, normalized such that
        the *integral* over the range is 1. Note that the sum of the
        histogram values will not be equal to 1 unless bins of unity
        width are chosen; it is not a probability *mass* function.

        Overrides the ``normed`` keyword if given.

    Returns
    -------
    hist : tensor
        The values of the histogram. See `density` and `weights` for a
        description of the possible semantics.
    bin_edges : tensor of dtype float
        Return the bin edges ``(length(hist)+1)``.


    See Also
    --------
    histogramdd, bincount, searchsorted, digitize, histogram_bin_edges

    Notes
    -----
    All but the last (righthand-most) bin is half-open.  In other words,
    if `bins` is::

      [1, 2, 3, 4]

    then the first bin is ``[1, 2)`` (including 1, but excluding 2) and
    the second ``[2, 3)``.  The last bin, however, is ``[3, 4]``, which
    *includes* 4.


    Examples
    --------
    >>> import mars.tensor as mt
    >>> mt.histogram([1, 2, 1], bins=[0, 1, 2, 3]).execute()
    (array([0, 2, 1]), array([0, 1, 2, 3]))
    >>> mt.histogram(mt.arange(4), bins=mt.arange(5), density=True).execute()
    (array([0.25, 0.25, 0.25, 0.25]), array([0, 1, 2, 3, 4]))
    >>> mt.histogram([[1, 2, 1], [1, 0, 1]], bins=[0,1,2,3]).execute()
    (array([1, 4, 1]), array([0, 1, 2, 3]))

    >>> a = mt.arange(5)
    >>> hist, bin_edges = mt.histogram(a, density=True)
    >>> hist.execute()
    array([0.5, 0. , 0.5, 0. , 0. , 0.5, 0. , 0.5, 0. , 0.5])
    >>> hist.sum().execute()
    2.4999999999999996
    >>> mt.sum(hist * mt.diff(bin_edges)).execute()
    1.0

    Automated Bin Selection Methods example, using 2 peak random data
    with 2000 points:

    >>> import matplotlib.pyplot as plt
    >>> rng = mt.random.RandomState(10)  # deterministic random data
    >>> a = mt.hstack((rng.normal(size=1000),
    ...                rng.normal(loc=5, scale=2, size=1000)))
    >>> _ = plt.hist(np.asarray(a), bins='auto')  # arguments are passed to np.histogram
    >>> plt.title("Histogram with 'auto' bins")
    Text(0.5, 1.0, "Histogram with 'auto' bins")
    >>> plt.show()

    """
    a, weights = _ravel_and_check_weights(a, weights)
    op = TensorHistogram(input=a, bins=bins, range=range,
                         weights=weights, density=density)
    return op(a, bins, range, weights)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import datetime
import json
import logging
import pickle
import sys
import traceback
import uuid
from concurrent.futures import ThreadPoolExecutor
from distutils.version import LooseVersion
from functools import lru_cache

from tornado import gen, web

from ..tensor.core import Indexes
from ..actors import new_client
from ..errors import GraphNotExists
from ..lib.tblib import pickling_support
from ..serialize.dataserializer import SerialType, CompressType
from ..utils import to_str, tokenize, numpy_dtype_from_descr_json, parse_readable_size
from .server import MarsWebAPI, MarsRequestHandler, register_web_handler

pickling_support.install()
_actor_client = new_client()

logger = logging.getLogger(__name__)


def _with_timeout(timeout, future):
    if timeout:
        try:
            yield gen.with_timeout(datetime.timedelta(seconds=timeout), future)
        except gen.TimeoutError:  # pragma: no cover
            raise TimeoutError
    else:
        yield future


class MarsApiRequestHandler(MarsRequestHandler):
    def set_default_headers(self):
        super().set_default_headers()
        self.set_header('Content-Type', 'application/json')

    def _dump_exception(self, exc_info, status_code=500):
        pickled_exc = pickle.dumps(exc_info)
        # return pickled exc_info for python client, and textual exc_info for web.
        self.write(json.dumps(dict(
            exc_info=base64.b64encode(pickled_exc).decode('ascii'),
            exc_info_text=traceback.format_exception(*exc_info),
        )))
        self.set_status(status_code)
        self.finish()

    @staticmethod
    @lru_cache(20)
    def _check_arrow_compatibility(client_version):
        import pyarrow
        client_version = tuple(LooseVersion(client_version or pyarrow.__version__).version[:2])
        server_version = tuple(LooseVersion(pyarrow.__version__).version[:2])
        return client_version == server_version

    def _handle_versions(self):
        args = {k: self.get_argument(k) for k in self.request.arguments}
        try:
            python_version = tuple(int(p) for p in args.pop('pyver').split('.')) \
                if 'pyver' in args else sys.version_info
            arrow_version = args.pop('arrow_version', None)
            pickle_protocol = int(args.pop('pickle_protocol', pickle.HIGHEST_PROTOCOL))
        except ValueError as ex:
            raise web.HTTPError(400, reason=f'Invalid version data: {ex}')
        if python_version[0] != sys.version_info[0]:
            raise web.HTTPError(400, reason='Python version not consistent')

        version_info = dict(python_version=python_version,
                            arrow_version=arrow_version,
                            arrow_compatible=self._check_arrow_compatibility(arrow_version),
                            pickle_protocol=min(pickle_protocol, pickle.HIGHEST_PROTOCOL))
        return version_info, args


class ApiEntryHandler(MarsApiRequestHandler):
    def get(self):
        self.write(dict(msg='Mars API Entry'))


class SessionsApiHandler(MarsApiRequestHandler):
    def get(self):
        self.write(json.dumps(self.web_api.get_sessions()))

    def post(self):
        versions, args = self._handle_versions()

        session_id = tokenize(str(uuid.uuid1()))
        self.web_api.create_session(session_id, **args)
        self.write(json.dumps(dict(
            session_id=session_id, arrow_compatible=versions['arrow_compatible'],
            pickle_protocol=versions['pickle_protocol'],
        )))


class SessionApiHandler(MarsApiRequestHandler):
    def get(self, session_id):
        if not self.web_api.has_session(session_id):
            raise web.HTTPError(404, 'Session doesn\'t not exists')

        name = self.get_argument('name', None)
        if name is None:
            versions, _ = self._handle_versions()

            self.write(json.dumps(dict(
                session_id=session_id, arrow_compatible=versions['arrow_compatible'],
                pickle_protocol=versions['pickle_protocol'],
            )))
        else:
            tileable_key = self.web_api.get_tileable_key_by_name(session_id, name)
            self.write(json.dumps(dict(tileable_key=tileable_key)))

    def delete(self, session_id):
        self.web_api.delete_session(session_id)


class GraphsApiHandler(MarsApiRequestHandler):
    def get(self, session_id):
        try:
            graph_states = self.web_api.get_tasks_info(session_id)
            tasks_dict = graph_states[session_id]['tasks']
            self.write(json.dumps(tasks_dict))
        except:  # noqa: E722
            self._dump_exception(sys.exc_info())

    def post(self, session_id):
        try:
            graph = self.get_argument('graph')
            target = self.get_argument('target').split(',')
            names = self.get_argument('names', default='').split(',')
            compose = bool(int(self.get_argument('compose', '1')))
        except web.MissingArgumentError as ex:
            self.write(json.dumps(dict(msg=str(ex))))
            raise web.HTTPError(400, reason='Argument missing')

        try:
            graph_key = tokenize(str(uuid.uuid4()))
            self.web_api.submit_graph(session_id, graph, graph_key, target, names=names, compose=compose)
            self.write(json.dumps(dict(graph_key=graph_key)))
        except:  # noqa: E722
            self._dump_exception(sys.exc_info())


class GraphApiHandler(MarsApiRequestHandler):
    _executor = ThreadPoolExecutor(1)

    @gen.coroutine
    def get(self, session_id, graph_key):
        from ..scheduler.utils import GraphState

        wait_timeout = self.get_argument('wait_timeout', None)
        if wait_timeout is not None:
            wait_timeout = float(wait_timeout)

        try:
            if wait_timeout:
                if wait_timeout <= 1e-6:
                    wait_timeout = None

                def _wait_fun():
                    web_api = MarsWebAPI(self._scheduler)
                    return web_api.wait_graph_finish(session_id, graph_key, wait_timeout)

                try:
                    yield from _with_timeout(wait_timeout, self._executor.submit(_wait_fun))
                    state = self.web_api.get_graph_state(session_id, graph_key)
                except TimeoutError:
                    state = GraphState.PREPARING
            else:
                state = self.web_api.get_graph_state(session_id, graph_key)
        except GraphNotExists:
            raise web.HTTPError(404, 'Graph not exists')

        resp = dict(state=state.value)
        if state == GraphState.FAILED:
            exc_info = self.web_api.get_graph_exc_info(session_id, graph_key)
            if exc_info is not None:
                resp['exc_info'] = base64.b64encode(pickle.dumps(exc_info)).decode('ascii')
                resp['exc_info_text'] = ''.join(traceback.format_exception(*exc_info))
        self.write(json.dumps(resp))

    def delete(self, session_id, graph_key):
        try:
            self.web_api.stop_graph(session_id, graph_key)
        except:  # noqa: E722
            self._dump_exception(sys.exc_info(), status_code=404)


class GraphDataApiHandler(MarsApiRequestHandler):
    _executor = ThreadPoolExecutor(1)

    @gen.coroutine
    def get(self, session_id, graph_key, tileable_key):
        data_type = self.get_argument('type', None)
        try:
            serial_type = SerialType(self.get_argument('serial_type', 'arrow'))
            pickle_protocol = int(self.get_argument('pickle_protocol', str(pickle.HIGHEST_PROTOCOL)))

            compressions_arg = self.get_argument('compressions', None)
            if compressions_arg:
                compressions_arg = [CompressType(s) for s in compressions_arg.split(',') if s]
            slices_arg = self.request.arguments.get('slices')
            if slices_arg:
                slices_arg = Indexes.from_json(json.loads(to_str(slices_arg[0]))).indexes
        except (TypeError, ValueError):
            raise web.HTTPError(403, 'Malformed encodings')

        if data_type:
            if data_type == 'nsplits':
                nsplits = self.web_api.get_tileable_nsplits(session_id, graph_key, tileable_key)
                self.write(json.dumps(nsplits))
            elif data_type == 'meta':
                meta = self.web_api.get_tileable_meta(session_id, graph_key, tileable_key)
                self.write(json.dumps(base64.b64encode(pickle.dumps(meta)).decode()))
            else:
                raise web.HTTPError(403, 'Unknown data type requests')
        else:

            def _fetch_fun():
                web_api = MarsWebAPI(self._scheduler)
                return web_api.fetch_data(session_id, tileable_key, index_obj=slices_arg,
                                          serial_type=serial_type, compressions=compressions_arg,
                                          pickle_protocol=pickle_protocol)

            data = yield self._executor.submit(_fetch_fun)
            self.write(data)

    def delete(self, session_id, graph_key, tileable_key):
        wait = int(self.get_argument('wait', '0'))
        self.web_api.delete_data(session_id, graph_key, tileable_key, wait=wait)


class OpLogHandler(MarsRequestHandler):
    _executor = ThreadPoolExecutor(1)

    @classmethod
    def _extract_size(cls, argument):
        if not argument:
            return
        if ',' not in argument and '=' not in argument:
            try:
                return int(argument)
            except ValueError:
                return int(parse_readable_size(argument)[0])
        else:
            ret = dict()
            for kv in argument.split(','):
                k, v = kv.split('=', 1)
                ret[k] = int(v)
            return ret

    @gen.coroutine
    def get(self, session_id, op_key):
        def _fetch_log():
            from ..context import DistributedContext

            offsets = self._extract_size(self.get_argument('offsets', None))
            sizes = self._extract_size(self.get_argument('sizes', None))

            ctx = DistributedContext(self._scheduler, session_id, is_distributed=True)
            log = ctx.fetch_tileable_op_logs(op_key, chunk_op_key_to_offsets=offsets,
                                             chunk_op_key_to_sizes=sizes)
            return log

        log_result = yield self._executor.submit(_fetch_log)
        self.write(json.dumps(log_result))


class WorkersApiHandler(MarsApiRequestHandler):
    _executor = ThreadPoolExecutor(1)

    def get(self):
        action = self.get_argument('action', None)
        if action == 'count':
            self.write(json.dumps(self.web_api.count_workers()))
        else:
            self.write(json.dumps(self.web_api.get_workers_meta()))

    @gen.coroutine
    def patch(self):
        action = self.get_argument('action', None)
        if action == 'count':
            wait = bool(int(self.get_argument('wait', '1')))
            timeout = int(self.get_argument('timeout', '0'))

            req_json = json.loads(self.request.body)
            count = req_json['new_scale']
            min_workers = req_json.get('min_workers')

            try:
                def _wait_fun():
                    web_api = MarsWebAPI(self._scheduler)
                    return web_api.rescale_workers(
                        count, min_workers=min_workers, wait=wait, timeout=timeout)

                yield from _with_timeout(timeout, self._executor.submit(_wait_fun))
            except NotImplementedError:
                self._dump_exception(sys.exc_info(), status_code=405)
            except:  # noqa: E722  # pragma: no cover  # pylint: disable=bare-except
                self._dump_exception(sys.exc_info())
        else:  # pragma: no cover
            raise web.HTTPError(405)


class MutableTensorApiHandler(MarsApiRequestHandler):
    def get(self, session_id, name):
        try:
            meta = self.web_api.get_mutable_tensor(session_id, name)
            self.write(json.dumps(meta))
        except:  # noqa: E722
            self._dump_exception(sys.exc_info())

    def post(self, session_id, name):
        try:
            action = self.get_argument('action', None)
            if action == 'create':
                req_json = json.loads(self.request.body.decode('ascii'))
                shape = req_json['shape']
                dtype = numpy_dtype_from_descr_json(req_json['dtype'])
                fill_value = req_json['fill_value']
                chunk_size = req_json['chunk_size']
                meta = self.web_api.create_mutable_tensor(session_id, name, shape, dtype,
                                                          fill_value=fill_value, chunk_size=chunk_size)
                self.write(json.dumps(meta))
            elif action == 'seal':
                info = self.web_api.seal(session_id, name)
                self.write(json.dumps(info))
            else:
                raise web.HTTPError(400, reason='Invalid argument')
        except:  # noqa: E722
            self._dump_exception(sys.exc_info())

    def put(self, session_id, name):
        try:
            payload_type = self.get_argument('payload_type', None)
            body = self.request.body
            self.web_api.write_mutable_tensor(session_id, name, payload_type, body)
        except:  # noqa: E722
            self._dump_exception(sys.exc_info())


register_web_handler('/api', ApiEntryHandler)
register_web_handler('/api/session', SessionsApiHandler)
register_web_handler('/api/worker', WorkersApiHandler)
register_web_handler('/api/session/(?P<session_id>[^/]+)', SessionApiHandler)
register_web_handler('/api/session/(?P<session_id>[^/]+)/graph', GraphsApiHandler)
register_web_handler('/api/session/(?P<session_id>[^/]+)/graph/(?P<graph_key>[^/]+)', GraphApiHandler)
register_web_handler('/api/session/(?P<session_id>[^/]+)/graph/(?P<graph_key>[^/]+)/data/(?P<tileable_key>[^/]+)',
                     GraphDataApiHandler)
register_web_handler('/api/session/(?P<session_id>[^/]+)/mutable-tensor/(?P<name>[^/]+)', MutableTensorApiHandler)
register_web_handler('/api/session/(?P<session_id>[^/]+)/op/(?P<op_key>[^/]+)/log', OpLogHandler)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from collections import defaultdict, OrderedDict
from datetime import datetime

import numpy as np
from bokeh.models import ColumnDataSource, Legend
from bokeh.plotting import figure

from ..utils import to_str
from .server import MarsRequestHandler, MarsWebAPI, get_jinja_env, \
    register_web_handler, register_bokeh_app

_jinja_env = get_jinja_env()
_worker_host_cache = dict()

TIMELINE_APP_NAME = 'bk_worker_timeline'
_TIMELINE_PALETTE = ['#0072B2', '#E69F00', '#F0E442', '#009E73', '#56B4E9',
                     '#D55E00', '#CC79A7', '#000000']

_dtype_datetime64_us = np.datetime64(datetime.now(), 'us').dtype


class WorkerListHandler(MarsRequestHandler):
    def get(self):
        workers_meta = self.web_api.get_workers_meta()
        for ep, meta in workers_meta.items():
            _worker_host_cache[ep] = meta['details']['host_name']

        template = _jinja_env.get_template('worker_pages/list.html')
        self.write_rendered(template, worker_metrics=workers_meta)


class WorkerHandler(MarsRequestHandler):
    def get(self, endpoint):
        workers_meta = self.web_api.get_workers_meta()
        for ep, meta in workers_meta.items():
            _worker_host_cache[ep] = meta['details']['host_name']

        progress_data = workers_meta[endpoint]['progress']
        progresses = dict()
        for session_data in progress_data.values():
            for g in session_data.values():
                if g['stage'] not in progresses:
                    progresses[g['stage']] = dict(operands=defaultdict(lambda: 0), total=0)
                progresses[g['stage']]['operands'][g['operands']] += 1
                progresses[g['stage']]['total'] += 1
        for k in progresses:
            operands = sorted(progresses[k]['operands'].items(), key=lambda p: p[0])
            progresses[k]['operands'] = ', '.join(f'{k} ({v})' for k, v in operands)

        template = _jinja_env.get_template('worker_pages/detail.html')
        self.write_rendered(
            template,
            endpoint=endpoint,
            worker_metrics=workers_meta[endpoint],
            progresses=progresses,
        )


class EventUpdater(object):
    def __init__(self):
        self.owner_to_ticker = OrderedDict()
        self.unfinished_to_indexes = dict()
        self.base_indexes = defaultdict(lambda: 0)

        self.min_left = datetime.max
        self.max_right = datetime.min
        self.min_bottom = 0xffffffff
        self.max_top = 0

    def update_events(self, event_list, cur_time=None):
        import pandas as pd
        from ..worker.events import ProcedureEventType

        cur_time = cur_time or time.time()
        cur_time_dt = datetime.fromtimestamp(cur_time)

        unfinished_to_indexes = self.unfinished_to_indexes
        owner_to_ticker = self.owner_to_ticker
        base_indexes = self.base_indexes

        event_dfs = dict()
        patches = defaultdict(list)
        new_unfinished = dict()

        lefts = defaultdict(list)
        rights = defaultdict(list)
        tops = defaultdict(list)
        bottoms = defaultdict(list)

        for event in event_list:
            ev_type = event.event_type

            try:
                ev_type, idx = unfinished_to_indexes.pop(event.event_id)
                patches[ev_type].append((idx, datetime.fromtimestamp(event.time_end)))
                continue
            except KeyError:
                pass

            try:
                ticker = owner_to_ticker[event.owner]
            except KeyError:
                ticker = owner_to_ticker[event.owner] = 1 + len(owner_to_ticker)

            bottoms[ev_type].append(ticker - 0.5)
            tops[ev_type].append(ticker + 0.5)
            lefts[ev_type].append(datetime.fromtimestamp(event.time_start))

            if event.time_end is None:
                new_unfinished[event.event_id] = (
                    ev_type, base_indexes[ev_type] + len(rights[ev_type])
                )
                rights[ev_type].append(cur_time_dt)
            else:
                rights[ev_type].append(datetime.fromtimestamp(event.time_end))

        for ev_type in ProcedureEventType.__members__.values():
            df = event_dfs[ev_type] = pd.DataFrame(dict(
                left=np.array(lefts[ev_type], dtype=_dtype_datetime64_us),
                right=np.array(rights[ev_type], dtype=_dtype_datetime64_us),
                top=np.array(tops[ev_type], dtype=np.float32),
                bottom=np.array(bottoms[ev_type], dtype=np.float32),
            ), columns=['left', 'right', 'top', 'bottom'])
            base_indexes[ev_type] += len(df)
            self.min_left = min(self.min_left, df.left.min())
            self.max_right = max(self.max_right, df.right.max())
            self.min_bottom = min(self.min_bottom, df.bottom.min())
            self.max_top = max(self.max_top, df.top.max())

        for ev_type, idx in unfinished_to_indexes.values():
            patches[ev_type].append((idx, cur_time_dt))
            self.max_right = max(self.max_right, cur_time_dt)
        unfinished_to_indexes.update(new_unfinished)

        return event_dfs, patches

    @property
    def x_range(self):
        if self.min_left > self.max_right:
            return None
        return self.min_left, self.max_right

    @property
    def y_range(self):
        if self.min_bottom > self.max_top:
            return None
        return self.min_bottom, self.max_top


class WorkerTimelineHandler(MarsRequestHandler):
    def get(self, endpoint):
        template = _jinja_env.get_template('worker_pages/timeline.html')
        worker_timeline_script = self.bokeh_server_document(
            TIMELINE_APP_NAME, arguments=dict(endpoint=endpoint))
        self.write_rendered(
            template,
            endpoint=endpoint,
            host_name=_worker_host_cache.get(endpoint, endpoint),
            worker_timeline_script=worker_timeline_script,
        )

    @staticmethod
    def timeline_app(doc, scheduler_ip=None):
        from ..worker.events import EventCategory, ProcedureEventType

        worker_ep = to_str(doc.session_context.request.arguments.get('endpoint')[0])
        web_api = MarsWebAPI(scheduler_ip)

        last_query_time = time.time()
        events = web_api.query_worker_events(
            worker_ep, EventCategory.PROCEDURE, time_end=last_query_time)

        updater = EventUpdater()
        data_sources = dict()

        dfs, _ = updater.update_events(events)

        p = figure(x_axis_type='datetime', plot_height=500, plot_width=800)
        renderers = []
        for idx, ev_type in enumerate(ProcedureEventType.__members__.values()):
            ds = data_sources[ev_type] = ColumnDataSource(dfs[ev_type])
            renderers.append(p.quad(left='left', right='right', top='top', bottom='bottom', source=ds,
                                    color=_TIMELINE_PALETTE[idx]))

        p.axis.major_tick_in = 0
        p.ygrid.grid_line_color = None
        p.outline_line_color = None

        old_owners = set()

        def _update_axes():
            x_range, y_range = updater.x_range, updater.y_range
            if x_range is None or y_range is None:
                return

            title_x_range = tuple(t.replace(microsecond=0) for t in x_range)
            p.title.text = f'Event time range: ({title_x_range[0]} - {title_x_range[1]})'

            if set(updater.owner_to_ticker.keys()) - old_owners:
                p.yaxis.ticker = list(range(1, 1 + len(updater.owner_to_ticker)))
                p.yaxis.major_label_overrides = dict((v, k) for k, v in updater.owner_to_ticker.items())
                old_owners.update(updater.owner_to_ticker.keys())

        _update_axes()

        legend_items = [('\0' + s, [r]) for s, r in zip(ProcedureEventType.__members__.keys(), renderers)]
        legend = Legend(items=legend_items, location='bottom_right',
                        orientation='horizontal', label_text_font_size='10px')
        p.add_layout(legend, 'below')

        doc.add_root(p)

        def _refresher():
            nonlocal last_query_time
            query_time = time.time()
            events = web_api.query_worker_events(
                worker_ep, EventCategory.PROCEDURE, time_start=last_query_time, time_end=query_time)
            last_query_time = query_time

            dfs, patches = updater.update_events(events)
            for ev_type, df in dfs.items():
                if len(df):
                    data_sources[ev_type].stream(df)

            for ev_type, patch in patches.items():
                if patch:
                    data_sources[ev_type].patch(dict(right=patch))

            _update_axes()

        doc.add_periodic_callback(_refresher, 5000)


register_web_handler('/worker', WorkerListHandler)
register_web_handler('/worker/(?P<endpoint>[^/]+)', WorkerHandler)
register_web_handler('/worker/(?P<endpoint>[^/]+)/timeline', WorkerTimelineHandler)

register_bokeh_app('/' + TIMELINE_APP_NAME, WorkerTimelineHandler.timeline_app)



# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import shutil
import sys
import tempfile

from .. import resource
from ..config import options
from ..utils import readable_size, calc_size_by_str
from .calc import CpuCalcActor, CudaCalcActor
from .custom_log import CustomLogFetchActor
from .dispatcher import DispatchActor
from .events import EventsActor
from .execution import ExecutionActor
from .prochelper import ProcessHelperActor
from .quota import QuotaActor, MemQuotaActor
from .status import StatusActor
from .storage import IORunnerActor, StorageManagerActor, SharedHolderActor, \
    InProcHolderActor, CudaHolderActor, DiskFileMergerActor
from .transfer import SenderActor, ReceiverManagerActor, ReceiverWorkerActor, ResultSenderActor
from .utils import WorkerClusterInfoActor


logger = logging.getLogger(__name__)


class WorkerService(object):
    def __init__(self, **kwargs):
        self._plasma_store = None

        self._storage_manager_ref = None
        self._shared_holder_ref = None
        self._task_queue_ref = None
        self._mem_quota_ref = None
        self._dispatch_ref = None
        self._events_ref = None
        self._status_ref = None
        self._execution_ref = None
        self._daemon_ref = None
        self._receiver_manager_ref = None

        self._cluster_info_ref = None
        self._cpu_calc_actors = []
        self._inproc_holder_actors = []
        self._inproc_io_runner_actors = []
        self._cuda_calc_actors = []
        self._cuda_holder_actors = []
        self._sender_actors = []
        self._receiver_actors = []
        self._spill_actors = []
        self._process_helper_actors = []
        self._custom_log_fetch_actors = []
        self._result_sender_ref = None
        self._file_merger_ref = None

        self._distributed = distributed = kwargs.pop('distributed', True)
        self._advertise_addr = kwargs.pop('advertise_addr', None)

        cuda_devices = kwargs.pop('cuda_devices', None) or os.environ.get('CUDA_VISIBLE_DEVICES')
        if not cuda_devices:
            self._n_cuda_process = 0
        else:
            cuda_devices = os.environ['CUDA_VISIBLE_DEVICES'] = ','.join(str(d) for d in cuda_devices)
            if cuda_devices:
                logger.info('Started Mars worker with CUDA cards %s', cuda_devices)
            self._n_cuda_process = resource.cuda_count()

        self._n_cpu_process = int(kwargs.pop('n_cpu_process', None) or resource.cpu_count())
        self._n_net_process = int(kwargs.pop('n_net_process', None) or '4')

        self._spill_dirs = kwargs.pop('spill_dirs', None)
        if self._spill_dirs:
            if isinstance(self._spill_dirs, str):
                from .utils import parse_spill_dirs
                self._spill_dirs = options.worker.spill_directory = parse_spill_dirs(self._spill_dirs)
            else:
                options.worker.spill_directory = self._spill_dirs
        else:
            self._spill_dirs = options.worker.spill_directory = []

        options.worker.disk_compression = kwargs.pop('disk_compression', None) \
            or options.worker.disk_compression
        options.worker.transfer_compression = kwargs.pop('transfer_compression', None) \
            or options.worker.transfer_compression
        options.worker.io_parallel_num = kwargs.pop('io_parallel_num', None) \
            or options.worker.io_parallel_num
        options.worker.recover_dead_process = not (kwargs.pop('disable_proc_recover', None)
                                                   or not options.worker.recover_dead_process)
        options.worker.write_shuffle_to_disk = kwargs.pop('write_shuffle_to_disk', None) \
            or options.worker.write_shuffle_to_disk

        min_cache_mem_size = kwargs.pop('min_cache_mem_size', None)
        min_cache_mem_size = min_cache_mem_size if min_cache_mem_size is not None \
            else options.worker.min_cache_mem_size
        options.worker.min_cache_mem_size = min_cache_mem_size

        plasma_limit = kwargs.pop('plasma_limit', None)
        plasma_limit = plasma_limit if plasma_limit is not None \
            else options.worker.plasma_limit
        options.worker.plasma_limit = plasma_limit

        if distributed and options.custom_log_dir is None:
            # gen custom_log_dir for distributed only
            options.custom_log_dir = tempfile.mkdtemp(prefix='mars-custom-log')
            self._clear_custom_log_dir = True
        else:
            self._clear_custom_log_dir = False

        self._total_mem = kwargs.pop('total_mem', None)
        if self._total_mem:
            os.environ['MARS_MEMORY_TOTAL'] = str(self._total_mem)
        self._cache_mem_size = kwargs.pop('cache_mem_size', None)
        self._cache_mem_scale = float(kwargs.pop('cache_mem_scale', None) or 1)
        self._soft_mem_limit = kwargs.pop('soft_mem_limit', None) or '80%'
        self._hard_mem_limit = kwargs.pop('hard_mem_limit', None) or '90%'
        self._ignore_avail_mem = kwargs.pop('ignore_avail_mem', None) or False
        self._min_mem_size = kwargs.pop('min_mem_size', None) or 128 * 1024 ** 2

        if sys.platform == 'win32':  # pragma: no cover
            raise NotImplementedError('Mars worker cannot start under Windows')
        elif sys.platform == 'darwin':
            default_plasma_dir = '/tmp'
        else:
            default_plasma_dir = '/dev/shm'
        options.worker.plasma_dir = kwargs.pop('plasma_dir', None) \
            or options.worker.plasma_dir or default_plasma_dir

        self._use_ext_plasma_dir = kwargs.pop('use_ext_plasma_dir', None) or False

        self._soft_quota_limit = self._soft_mem_limit

        self._calc_memory_limits()

        if kwargs:  # pragma: no cover
            raise TypeError(f'Keyword arguments {kwargs!r} cannot be recognized.')

    @property
    def n_process(self):
        return 1 + self._n_cpu_process + self._n_cuda_process + self._n_net_process \
               + (1 if self._spill_dirs else 0)

    @staticmethod
    def _get_plasma_size():
        fd = os.open(options.worker.plasma_dir, os.O_RDONLY)
        stats = os.fstatvfs(fd)
        os.close(fd)
        size = stats.f_bsize * stats.f_bavail
        # keep some safety margin for allocator fragmentation
        return 8 * size // 10

    def _calc_memory_limits(self):
        mem_stats = resource.virtual_memory()

        if self._total_mem:
            self._total_mem = calc_size_by_str(self._total_mem, mem_stats.total)
        else:
            self._total_mem = mem_stats.total

        self._min_mem_size = calc_size_by_str(self._min_mem_size, self._total_mem)
        self._hard_mem_limit = calc_size_by_str(self._hard_mem_limit, self._total_mem)

        raw_cache_mem_size = self._cache_mem_size = \
            calc_size_by_str(self._cache_mem_size, self._total_mem)
        if self._cache_mem_size is None:
            raw_cache_mem_size = self._cache_mem_size = mem_stats.free // 2

        plasma_size = self._get_plasma_size()
        if plasma_size is not None:
            self._cache_mem_size = min(plasma_size, self._cache_mem_size)
        self._cache_mem_size = int(self._cache_mem_size * self._cache_mem_scale)

        self._soft_mem_limit = calc_size_by_str(self._soft_mem_limit, self._total_mem)
        actual_used = self._total_mem - mem_stats.available
        if self._ignore_avail_mem:
            self._soft_quota_limit = self._soft_mem_limit
        else:
            used_cache_size = 0 if self._use_ext_plasma_dir else self._cache_mem_size
            self._soft_quota_limit = self._soft_mem_limit - used_cache_size - actual_used
            if self._soft_quota_limit < self._min_mem_size:
                raise MemoryError(
                    f'Memory not enough. soft_limit={readable_size(self._soft_mem_limit)}, '
                    f'cache_limit={readable_size(self._cache_mem_size)}, '
                    f'used={readable_size(actual_used)}')

        if options.worker.min_cache_mem_size:
            min_cache_mem_size = calc_size_by_str(options.worker.min_cache_mem_size, self._total_mem)
            if min(min_cache_mem_size, raw_cache_mem_size) > self._cache_mem_size:
                raise MemoryError(f'Cache memory size ({self._cache_mem_size}) smaller than '
                                  f'minimal size ({min_cache_mem_size}), worker cannot start')

        logger.info('Setting soft limit to %s.', readable_size(self._soft_quota_limit))

    def start_plasma(self):
        from pyarrow import plasma
        self._plasma_store = plasma.start_plasma_store(
            self._cache_mem_size, plasma_directory=options.worker.plasma_dir)
        options.worker.plasma_socket, _ = self._plasma_store.__enter__()

    def start(self, endpoint, pool, discoverer=None, process_start_index=0):
        distributed = self._distributed
        # create plasma key mapper
        from .storage import PlasmaKeyMapActor
        pool.create_actor(PlasmaKeyMapActor, uid=PlasmaKeyMapActor.default_uid())

        # create WorkerClusterInfoActor
        self._cluster_info_ref = pool.create_actor(
            WorkerClusterInfoActor, discoverer, distributed=distributed,
            uid=WorkerClusterInfoActor.default_uid())

        # create DiskFileMergerActor
        if options.worker.filemerger.enabled and options.worker.spill_directory:
            self._file_merger_ref = pool.create_actor(
                DiskFileMergerActor, uid=DiskFileMergerActor.default_uid())

        if distributed:
            # create process daemon
            from .daemon import WorkerDaemonActor
            actor_holder = self._daemon_ref = pool.create_actor(
                WorkerDaemonActor, uid=WorkerDaemonActor.default_uid())

            # create StatusActor
            if ':' not in self._advertise_addr:
                self._advertise_addr += ':' + endpoint.rsplit(':', 1)[-1]

            self._status_ref = pool.create_actor(
                StatusActor, self._advertise_addr, uid=StatusActor.default_uid())
        else:
            # create StatusActor
            self._status_ref = pool.create_actor(
                StatusActor, endpoint, with_gpu=self._n_cuda_process > 0, uid=StatusActor.default_uid())

            actor_holder = pool

        if self._ignore_avail_mem:
            # start a QuotaActor instead of MemQuotaActor to avoid memory size detection
            # for debug purpose only, DON'T USE IN PRODUCTION
            self._mem_quota_ref = pool.create_actor(
                QuotaActor, self._soft_mem_limit, uid=MemQuotaActor.default_uid())
        else:
            self._mem_quota_ref = pool.create_actor(
                MemQuotaActor, self._soft_quota_limit, self._hard_mem_limit, uid=MemQuotaActor.default_uid())

        # create StorageManagerActor
        self._storage_manager_ref = pool.create_actor(
            StorageManagerActor, uid=StorageManagerActor.default_uid())
        # create SharedHolderActor
        start_timeout = int(os.environ.get('MARS_SHARED_HOLDER_START_TIMEOUT', None) or 60)
        start_future = pool.create_actor(
            SharedHolderActor, self._cache_mem_size, uid=SharedHolderActor.default_uid(), wait=False)
        self._shared_holder_ref = start_future.result(start_timeout)
        # create DispatchActor
        self._dispatch_ref = pool.create_actor(DispatchActor, uid=DispatchActor.default_uid())
        # create EventsActor
        self._events_ref = pool.create_actor(EventsActor, uid=EventsActor.default_uid())
        # create ReceiverNotifierActor
        self._receiver_manager_ref = pool.create_actor(ReceiverManagerActor, uid=ReceiverManagerActor.default_uid())
        # create ExecutionActor
        self._execution_ref = pool.create_actor(ExecutionActor, uid=ExecutionActor.default_uid())

        # create CpuCalcActor and InProcHolderActor
        if not distributed:
            self._n_cpu_process = pool.cluster_info.n_process - 1 - process_start_index

        for cpu_id in range(self._n_cpu_process):
            uid = f'w:{cpu_id + 1}:mars-cpu-calc'
            actor = actor_holder.create_actor(CpuCalcActor, uid=uid)
            self._cpu_calc_actors.append(actor)

            uid = f'w:{cpu_id + 1}:mars-inproc-holder'
            actor = actor_holder.create_actor(InProcHolderActor, uid=uid)
            self._inproc_holder_actors.append(actor)

            actor = actor_holder.create_actor(
                IORunnerActor, dispatched=False, uid=IORunnerActor.gen_uid(cpu_id + 1))
            self._inproc_io_runner_actors.append(actor)

        start_pid = 1 + self._n_cpu_process

        stats = resource.cuda_card_stats() if self._n_cuda_process else []
        for cuda_id, stat in enumerate(stats):
            for thread_no in range(options.worker.cuda_thread_num):
                uid = f'w:{start_pid + cuda_id}:mars-cuda-calc-{cuda_id}-{thread_no}'
                actor = actor_holder.create_actor(CudaCalcActor, uid=uid)
                self._cuda_calc_actors.append(actor)

            uid = f'w:{start_pid + cuda_id}:mars-cuda-holder-{cuda_id}'
            actor = actor_holder.create_actor(
                CudaHolderActor, stat.fb_mem_info.total, device_id=stat.index, uid=uid)
            self._cuda_holder_actors.append(actor)

            actor = actor_holder.create_actor(
                IORunnerActor, dispatched=False, uid=IORunnerActor.gen_uid(start_pid + cuda_id))
            self._inproc_io_runner_actors.append(actor)

        start_pid += self._n_cuda_process

        if distributed:
            # create SenderActor and ReceiverActor
            for sender_id in range(self._n_net_process):
                uid = f'w:{start_pid + sender_id}:mars-sender-{sender_id}'
                actor = actor_holder.create_actor(SenderActor, uid=uid)
                self._sender_actors.append(actor)

        if options.custom_log_dir is not None:
            for custom_log_fetch_id in range(self._n_net_process):
                uid = f'w:{start_pid + custom_log_fetch_id}:mars-custom-log-fetch'
                actor = actor_holder.create_actor(CustomLogFetchActor, uid=uid)
                self._custom_log_fetch_actors.append(actor)

        # Mutable requires ReceiverActor (with ClusterSession)
        for receiver_id in range(2 * self._n_net_process):
            uid = f'w:{start_pid + receiver_id // 2}:mars-receiver-{receiver_id}'
            actor = actor_holder.create_actor(ReceiverWorkerActor, uid=uid)
            self._receiver_actors.append(actor)

        # create ProcessHelperActor
        for proc_id in range(pool.cluster_info.n_process - process_start_index):
            uid = f'w:{proc_id}:mars-process-helper'
            actor = actor_holder.create_actor(ProcessHelperActor, uid=uid)
            self._process_helper_actors.append(actor)

        # create ResultSenderActor
        self._result_sender_ref = pool.create_actor(ResultSenderActor, uid=ResultSenderActor.default_uid())

        # create SpillActor
        start_pid = pool.cluster_info.n_process - 1
        if options.worker.spill_directory:
            for spill_id in range(len(options.worker.spill_directory)):
                uid = f'w:{start_pid}:mars-global-io-runner-{spill_id}'
                actor = actor_holder.create_actor(IORunnerActor, uid=uid)
                self._spill_actors.append(actor)

        # worker can be registered when everything is ready
        self._status_ref.enable_status_upload(_tell=True)

    def handle_process_down(self, pool, proc_indices):
        proc_id_to_pid = dict((proc_id, pool.processes[proc_id].pid) for proc_id in proc_indices)
        exit_codes = [pool.processes[proc_id].exitcode for proc_id in proc_indices]
        logger.warning('Process %r halt, exitcodes=%r. Trying to recover.', proc_id_to_pid, exit_codes)
        for proc_id in proc_indices:
            pool.restart_process(proc_id)
        self._daemon_ref.handle_process_down(proc_indices)

    def stop(self):
        try:
            destroy_futures = []
            for actor in (self._cpu_calc_actors + self._sender_actors + self._inproc_holder_actors
                          + self._inproc_io_runner_actors + self._cuda_calc_actors
                          + self._cuda_holder_actors + self._receiver_actors + self._spill_actors
                          + self._process_helper_actors):
                if actor and actor.ctx:
                    destroy_futures.append(actor.destroy(wait=False))

            if self._result_sender_ref:
                destroy_futures.append(self._result_sender_ref.destroy(wait=False))
            if self._status_ref:
                destroy_futures.append(self._status_ref.destroy(wait=False))
            if self._shared_holder_ref:
                destroy_futures.append(self._shared_holder_ref.destroy(wait=False))
            if self._storage_manager_ref:
                destroy_futures.append(self._storage_manager_ref.destroy(wait=False))
            if self._events_ref:
                destroy_futures.append(self._events_ref.destroy(wait=False))
            if self._dispatch_ref:
                destroy_futures.append(self._dispatch_ref.destroy(wait=False))
            if self._execution_ref:
                destroy_futures.append(self._execution_ref.destroy(wait=False))
            [f.result(5) for f in destroy_futures]

            if self._clear_custom_log_dir:
                custom_dir = options.custom_log_dir
                shutil.rmtree(custom_dir, ignore_errors=True)
                options.custom_log_dir = None
        finally:
            if not options.vineyard.enabled:
                self._plasma_store.__exit__(None, None, None)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ..datasource import tensor as astensor
from .core import TensorComplexFFTMixin, validate_fft, TensorStandardFFT


class TensorIFFT(TensorStandardFFT, TensorComplexFFTMixin):
    _op_type_ = OperandDef.IFFT

    def __init__(self, n=None, axis=-1, norm=None, dtype=None, **kw):
        super().__init__(_n=n, _axis=axis, _norm=norm, _dtype=dtype, **kw)


def ifft(a, n=None, axis=-1, norm=None):
    """
    Compute the one-dimensional inverse discrete Fourier Transform.

    This function computes the inverse of the one-dimensional *n*-point
    discrete Fourier transform computed by `fft`.  In other words,
    ``ifft(fft(a)) == a`` to within numerical accuracy.
    For a general description of the algorithm and definitions,
    see `mt.fft`.

    The input should be ordered in the same way as is returned by `fft`,
    i.e.,

    * ``a[0]`` should contain the zero frequency term,
    * ``a[1:n//2]`` should contain the positive-frequency terms,
    * ``a[n//2 + 1:]`` should contain the negative-frequency terms, in
      increasing order starting from the most negative frequency.

    For an even number of input points, ``A[n//2]`` represents the sum of
    the values at the positive and negative Nyquist frequencies, as the two
    are aliased together. See `numpy.fft` for details.

    Parameters
    ----------
    a : array_like
        Input tensor, can be complex.
    n : int, optional
        Length of the transformed axis of the output.
        If `n` is smaller than the length of the input, the input is cropped.
        If it is larger, the input is padded with zeros.  If `n` is not given,
        the length of the input along the axis specified by `axis` is used.
        See notes about padding issues.
    axis : int, optional
        Axis over which to compute the inverse DFT.  If not given, the last
        axis is used.
    norm : {None, "ortho"}, optional
        Normalization mode (see `numpy.fft`). Default is None.

    Returns
    -------
    out : complex Tensor
        The truncated or zero-padded input, transformed along the axis
        indicated by `axis`, or the last one if `axis` is not specified.

    Raises
    ------
    IndexError
        If `axes` is larger than the last axis of `a`.

    See Also
    --------
    mt.fft : An introduction, with definitions and general explanations.
    fft : The one-dimensional (forward) FFT, of which `ifft` is the inverse
    ifft2 : The two-dimensional inverse FFT.
    ifftn : The n-dimensional inverse FFT.

    Notes
    -----
    If the input parameter `n` is larger than the size of the input, the input
    is padded by appending zeros at the end.  Even though this is the common
    approach, it might lead to surprising results.  If a different padding is
    desired, it must be performed before calling `ifft`.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.fft.ifft([0, 4, 0, 0]).execute()
    array([ 1.+0.j,  0.+1.j, -1.+0.j,  0.-1.j])

    Create and plot a band-limited signal with random phases:

    >>> import matplotlib.pyplot as plt
    >>> t = mt.arange(400)
    >>> n = mt.zeros((400,), dtype=complex)
    >>> n[40:60] = mt.exp(1j*mt.random.uniform(0, 2*mt.pi, (20,)))
    >>> s = mt.fft.ifft(n)
    >>> plt.plot(t.execute(), s.real.execute(), 'b-', t.execute(), s.imag.execute(), 'r--')
    ...
    >>> plt.legend(('real', 'imaginary'))
    ...
    >>> plt.show()

    """
    a = astensor(a)
    validate_fft(a, axis, norm)
    op = TensorIFFT(n=n, axis=axis, norm=norm, dtype=np.dtype(np.complex_))
    return op(a)
